/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: dmonmsg.h
 */

#include <sys/ioctl.h>
#include <common.h>

#ifndef EVMS_DMONMSG_H_INCLUDED
#define EVMS_DMONMSG_H_INCLUDED 1

#define EVMS_DMONMSG_MAJOR_VERSION	1
#define EVMS_DMONMSG_MINOR_VERSION	0
#define EVMS_DMONMSG_PATCH_LEVEL	0
				
typedef enum {
	MSG_GET_VERSION = 1,

	MSG_LOCK_ENGINE,
	MSG_UNLOCK_ENGINE,

	MSG_CHKNOD,
	MSG_MKNOD,

	MSG_DISCOVER_DISKS,

	MSG_GET_DEV_NODE_NAMES,

	MSG_OPEN,
	MSG_READ,
	MSG_WRITE,
	MSG_IOCTL,
	MSG_CLOSE,
	MSG_UNLINK,

	MSG_IS_MOUNTED,
	MSG_CHECK_OPEN_STATUS,

	MSG_COPY,

	MSG_WRITE_LOG,
	MSG_USER_MSG,

	MSG_DM_CMD,

	MSG_EXECVP,
	MSG_EXECVP_STDIN_IO,
	MSG_EXECVP_STDOUT_IO,
	MSG_EXECVP_STDERR_IO,

	MSG_INVALID_CMD		= 0xffffffff
} msg_cmd_t;


#define COMMAND_RESPONSE	0x80000000
#define COMMAND_MASK		~COMMAND_RESPONSE


static inline char * msg_name(msg_cmd_t cmd) {

	switch (cmd & ~COMMAND_RESPONSE) {
	case MSG_GET_VERSION:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_GET_VERSION response";
		} else {
			return "MSG_GET_VERSION request";
		}
		break;
	case MSG_LOCK_ENGINE:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_LOCK_ENGINE response";
		} else {
			return "MSG_LOCK_ENGINE request";
		}
		break;
	case MSG_UNLOCK_ENGINE:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_UNLOCK_ENGINE response";
		} else {
			return "MSG_UNLOCK_ENGINE request";
		}
		break;
	case MSG_CHKNOD:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_CHKNOD response";
		} else {
			return "MSG_CHKNOD request";
		}
		break;
	case MSG_MKNOD:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_MKNOD response";
		} else {
			return "MSG_MKNOD request";
		}
		break;
	case MSG_DISCOVER_DISKS:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_DISCOVER_DISKS response";
		} else {
			return "MSG_DISCOVER_DISKS request";
		}
		break;
	case MSG_GET_DEV_NODE_NAMES:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_GET_DEV_NODE_NAMES response";
		} else {
			return "MSG_GET_DEV_NODE_NAMES request";
		}
		break;
	case MSG_OPEN:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_OPEN response";
		} else {
			return "MSG_OPEN request";
		}
		break;
	case MSG_READ:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_READ response";
		} else {
			return "MSG_READ request";
		}
		break;
	case MSG_WRITE:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_WRITE response";
		} else {
			return "MSG_WRITE request";
		}
		break;
	case MSG_IOCTL:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_IOCTL response";
		} else {
			return "MSG_IOCTL request";
		}
		break;
	case MSG_CLOSE:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_CLOSE response";
		} else {
			return "MSG_CLOSE request";
		}
		break;
	case MSG_UNLINK:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_UNLINK response";
		} else {
			return "MSG_UNLINK request";
		}
		break;
	case MSG_IS_MOUNTED:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_IS_MOUNTED response";
		} else {
			return "MSG_IS_MOUNTED request";
		}
		break;
	case MSG_CHECK_OPEN_STATUS:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_CHECK_OPEN_STATUS response";
		} else {
			return "MSG_CHECK_OPEN_STATUS request";
		}
		break;
	case MSG_COPY:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_COPY response";
		} else {
			return "MSG_COPY request";
		}
		break;
	case MSG_WRITE_LOG:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_WRITE_LOG response";
		} else {
			return "MSG_WRITE_LOG request";
		}
		break;
	case MSG_USER_MSG:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_USER_MSG response";
		} else {
			return "MSG_USER_MSG request";
		}
		break;
	case MSG_DM_CMD:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_DM_CMD response";
		} else {
			return "MSG_DM_CMD request";
		}
		break;
	case MSG_EXECVP:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_EXECVP response";
		} else {
			return "MSG_EXECVP request";
		}
		break;
	case MSG_EXECVP_STDIN_IO:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_EXECVP_STDIN_IO response";
		} else {
			return "MSG_EXECVP_STDIN_IO request";
		}
		break;
	case MSG_EXECVP_STDOUT_IO:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_EXECVP_STDOUT_IO response";
		} else {
			return "MSG_EXECVP_STDOUT_IO request";
		}
		break;
	case MSG_EXECVP_STDERR_IO:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_EXECVP_STDERR_IO response";
		} else {
			return "MSG_EXECVP_STDERR_IO request";
		}
		break;
	case MSG_INVALID_CMD:
		if (cmd & COMMAND_RESPONSE) {
			return "MSG_INVALID_CMD response";
		} else {
			return "MSG_INVALID_CMD request";
		}
		break;
	default:
		if (cmd & COMMAND_RESPONSE) {
			return "UNKNOWN response";
		} else {
			return "UNKNOWN request";
		}
		break;
	}
}


/* There are no parameters for MSG_GET_VERSION. */
typedef struct msg_get_version_s {
} msg_get_version_t;


typedef struct msg_get_version_response_s {
	evms_version_t  version;
} msg_get_version_response_t;


typedef struct msg_lock_engine_s {
	u_int32_t       mode;
} msg_lock_engine_t;


typedef struct msg_lock_engine_response_s {
	u_int32_t       status;
} msg_lock_engine_response_t;


/* There are no parameters for MSG_UNLOCK_ENGINE. */
typedef struct msg_unlock_engine_s {
} msg_unlock_engine_t;


typedef struct msg_unlock_engine_response_s {
	int32_t         status;
} msg_unlock_engine_response_t;


typedef struct msg_chknod_s {
	u_int32_t       major;
	u_int32_t       minor;
	char            name[0];
} msg_chknod_t;


typedef struct msg_chknod_response_s {
	int32_t         status;
} msg_chknod_response_t;


typedef struct msg_mknod_s {
	u_int32_t       major;
	u_int32_t       minor;
	char            name[0];
} msg_mknod_t;


typedef struct msg_mknod_response_s {
	int32_t         status;
} msg_mknod_response_t;


typedef struct msg_discover_disks_s {
} msg_discover_disks_t;


typedef struct msg_discover_disks_response_s {
	int32_t             status;
	u_int32_t           num_disks;
	storage_object_t    disks[0];
} msg_discover_disks_response_t;


typedef struct msg_get_dev_node_names_s {
	char            dir_name[0];
} msg_get_dev_node_names_t;


typedef struct msg_get_dev_node_names_response_s {
	int32_t         status;
	u_int32_t       count;
	char            names[0];
} msg_get_dev_node_names_response_t;


typedef struct msg_open_s {
	u_int32_t       flags;
	u_int32_t       mode;
	char            name[0];
} msg_open_t;


typedef struct msg_open_response_s {
	int32_t         status;
	int32_t         fd;
} msg_open_response_t;


typedef struct msg_read_s {
	u_int64_t       offset;
	u_int32_t       length;
	u_int32_t       fd;
} msg_read_t;


typedef struct msg_read_response_s {
	int32_t         bytes_read;
	int32_t         status;
	unsigned char   data[0];
} msg_read_response_t;


typedef struct msg_write_s {
	u_int64_t       offset;
	u_int32_t       length;
	u_int32_t       fd;
	unsigned char   data[0];
} msg_write_t;


typedef struct msg_write_response_s {
	int32_t         bytes_written;
	int32_t         status;
} msg_write_response_t;


/*
 * ioctl processors use the same memory buffer for input and output.
 * Rather than uses separate buffers for the message and response,
 * it is easier to use one buffer for both and avoid the memory copies.
 */
typedef struct msg_ioctl_s {
	u_int64_t       cmd;
	u_int32_t       fd;
	int32_t         status;
	unsigned char   arg[0];
} msg_ioctl_t;


typedef struct msg_close_s {
	u_int32_t       fd;
} msg_close_t;


typedef struct msg_close_response_s {
	int32_t         status;
} msg_close_response_t;


typedef struct msg_unlink_s {
	char            file_name[0];
} msg_unlink_t;


typedef struct msg_unlink_response_s {
	u_int32_t       status;
} msg_unlink_response_t;


typedef struct msg_is_mounted_s {
	char            volume_name[0];
} msg_is_mounted_t;


typedef struct msg_is_mounted_response_s {
	int32_t         status;
	u_int32_t       mounted;	/* TRUE or FALSE */
	char            mount_name[0];
} msg_is_mounted_response_t;


typedef struct msg_check_open_status_s {
	char            volume_name[0];
} msg_check_open_status_t;


typedef struct msg_check_open_status_response_s {
	int32_t         status;
	int32_t         opened;		/* TRUE or FALSE */
} msg_check_open_status_response_t;


typedef struct msg_copy_s {
	u_int64_t	src_offset;
	u_int64_t	trg_offset;
	u_int64_t	len;
	u_int64_t	context;
	int32_t		src_fd;
	int32_t		trg_fd;
	int32_t		chunk_size;
} msg_copy_t;


typedef struct msg_copy_response_s {
	u_int64_t	context;
	u_int64_t	chunks_copied;
	u_int32_t	status;
} msg_copy_response_t;


typedef struct msg_write_log_s {
	char            msg[0];
} msg_write_log_t;


typedef struct msg_write_log_response_s {
	int32_t         status;
} msg_write_log_response_t;


typedef struct msg_user_msg_s {
	char            msg[0];
} msg_user_msg_t;


typedef struct msg_user_msg_response_s {
	int32_t         status;
} msg_user_msg_response_t;


typedef struct msg_dm_cmd_s {
	u_int64_t       cmd;
	u_int32_t	size;
	char            ioctl_data[0];
} msg_dm_cmd_t;


typedef struct msg_dm_cmd_response_s {
	u_int64_t       status;         // Same size as "cmd" in msg_dm_cmd_t
					// so that the same buffer can be used
					// for request and response.
	u_int32_t	size;
	char            ioctl_data[0];
} msg_dm_cmd_response_t;


typedef struct msg_execvp_s {
	char            flat_argv[0];
} msg_execvp_t;

/*
 * Standard ASCII code used to mark the end of text.
 * We can't use an addtional terminating nul, because any of the argv
 * strings could be an empty string which would cause a double nul
 * in the data which could be mistaken for the end of the argv strings.
 */
#define END_OF_TEXT	'\x03'

typedef struct msg_execvp_response_s {
	u_int32_t	rc;
	u_int32_t       status;
} msg_execvp_response_t;


typedef struct msg_execvp_io_s {
	char            data[0];
} msg_execvp_io_t;


typedef struct msg_execvp_io_response_s {
	u_int32_t       status;
} msg_execvp_io_response_t;

#endif
