/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: remote.h
 *
 */

/*
 * Global functions exported by remote.c
 */

/* Internal functions to share with other Engine source modules */

extern int remote_verify_version(void);

extern int remote_lock_engine(void);

extern int remote_unlock_engine(void);

extern int remote_hasa_dev_node(ece_nodeid_t * node,
				char         * name,
				u_int32_t      major,
				u_int32_t      minor);

extern int remote_ensure_dev_node(ece_nodeid_t * node,
				  char         * name,
				  u_int32_t      major,
				  u_int32_t      minor);

extern int remote_discover_disks(ece_nodeid_t       * node,
				 u_int32_t          * num_disks,
				 storage_object_t * * info);

extern int remote_get_dev_node_names(ece_nodeid_t * node,
				     char         * dir_name,
				     char     * * * name_list,
				     int          * count);

extern int remote_open(ece_nodeid_t * node,
		       char         * name,
		       int            flags,
		       u_int32_t      mode,
		       int          * fd);

extern int remote_read(ece_nodeid_t * node,
		       int            fd,
		       u_int64_t      offset,
		       u_int32_t      length,
		       void *         buffer,
		       int32_t      * bytes_read);

extern int remote_write(ece_nodeid_t * node,
			int            fd,
			u_int64_t      offset,
			u_int32_t      length,
			void *         buffer,
			int32_t      * bytes_written);

extern int remote_ioctl(ece_nodeid_t  * node,
			int            fd,
			unsigned long  cmd,
			void         * arg);

extern int remote_close(ece_nodeid_t * node,
			int            fd);

extern int remote_unlink(ece_nodeid_t  * node,
			 char          * file_name);

extern int remote_is_mounted(ece_nodeid_t * node,
			     char         * volume_name,
			     BOOLEAN      * mounted,
			     char       * * mount_name);

extern int remote_check_open_status(ece_nodeid_t * node,
				    char         * volume_name,
				    u_int32_t    * opens);

extern int remote_copy(ece_nodeid_t  * node,
		       int             src_fd,
		       u_int64_t       src_offset,
		       int             trg_fd,
		       u_int64_t       trg_offset,
		       u_int64_t       len,
		       int32_t         chunk_size,
		       void          * context,
		       copy_callback_t callback);

extern int remote_dm_command(ece_nodeid_t  * node,
		             void          * dmi,
		             unsigned long   data_size,
		             unsigned long   command);

extern void engine_router(ece_msg_t * msg);

extern pid_t remote_fork_and_execvp(ece_nodeid_t * node,
				    char         * argv[],
				    int            stdin_pipe[2],
				    int            stdout_pipe[2],
				    int            stderr_pipe[2]);

extern void msg_execvp_stdin_io(const ece_msg_t * msg);

extern void msg_execvp_stdout_io(const ece_msg_t * msg);

extern void msg_execvp_stderr_io(const ece_msg_t * msg);

