/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_regions.h
 */

#ifndef _AIX_REGIONS_H_
#define _AIX_REGIONS_H_

extern int aix_check_lv_name( unsigned char * lv_name );
extern storage_object_t * aix_get_freespace_region( dlist_t selected_objects );
extern int aix_deallocate_region(	storage_object_t	* region,
					boolean			delete_evms_objects );
extern storage_object_t * aix_allocate_region(	lv_entries		* lv,
						char			* lv_name,
						unsigned int		pp_size );
extern void aix_clear_lv_entry( lv_entries * lv );
extern lv_entries * aix_initialize_new_lv_entry(storage_container_t	* container,
						aix_region_options_t	* region_opts );
extern int aix_create_freespace_region( storage_container_t * container );
extern int aix_update_freespace_region( storage_container_t * container );
extern int aix_build_lp_maps( storage_container_t * container, storage_object_t * mirror_object, storage_object_t * mirror_object2,int expand );
extern int aix_discover_regions( void );
extern int aix_export_regions( dlist_t regions );
extern int aix_allocate_partitions_to_region(	storage_object_t	* region,
						storage_container_t	* container,
						aix_region_options_t	* region_opts,
                        unsigned short            operation);
extern int aix_expand_allocate_partitions_to_region(	storage_object_t	* region,
						storage_container_t	* container,
						aix_region_options_t	* region_opts,
                        unsigned short            operation);
extern int aix_deallocate_partitions_from_region( storage_object_t * region, int count );

extern int aix_check_lv_size(	u_int32_t	* add_size,    u_int32_t	pe_size );

extern int aix_check_physical_layout(storage_object_t  * object,
                              int               * AllOnOnePV);

extern int aix_add_mirrors_to_object(storage_object_t     * object,
                              storage_object_t     * mirror_object,
                              storage_object_t     * mirror_object2,
                              int                    AllonOnePV,
                              storage_container_t  * container);

extern int aix_create_mirror_partitions(storage_object_t    * region, 
                                 storage_object_t    * mirror_object, 
                                 storage_container_t * container);

extern int aix_allocate_mirror_partitions_to_region( storage_object_t	    * region,
                                              storage_object_t      * mirror_object,
					                          storage_container_t	* container,
                                              int                     num_lps,
                                              int                     AllOnOnePV);


extern storage_object_t * aix_pv_from_object(storage_container_t * container, storage_object_t * object, int * pv_num);

extern int aix_connect_mapped_lvs(storage_container_t *container, 
                                  storage_object_t * object, 
								  storage_object_t * mirror_object, 
								  storage_object_t * mirror_object2, 
								  int AllOnOnePV); 

extern int aix_get_free_pps_for_object(storage_object_t  *object,
		            					aix_region_data_t * r_data);
extern storage_object_t * aix_get_mirror_object(aix_container_data_t *c_data, aix_region_data_t * r_data);
extern int  aix_split_mirror_region(storage_object_t * object, option_array_t * options);

extern int aix_split_partitions_on_region(	storage_object_t	* region,
			                    	storage_container_t	* container,
     				                aix_region_options_t	* region_opts,
									storage_object_t    * new_region);
#endif
