/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aixregmgr.c
 *
 * Description: This file contains all of the required engine-plugin APIs
 *              for the AIX region manager. This is also where all AIX global
 *              data is declared.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"



// Global variables
engine_functions_t              * EngFncs;		    // The Engine's internal API set.
plugin_record_t                 * my_plugin_record;		// Used to identify containers and regions as owned by AIX.
dlist_t                           aix_container_list;	// List of real containers.
dlist_t                   aix_mirror_objects;	// List of acceptable mirrors


/* aix_setup_evms_plugin
 *
 *	This function gets called shortly after the plugin is loaded by the
 *	Engine. It performs all tasks that are necessary before the initial
 *	discovery pass.
 */

#if 0
static int can_be_added(ADDRESS object,
			TAG     object_tag,
			uint    object_size,
			ADDRESS object_handle,
			ADDRESS parameters) {

	storage_object_t * mirror_candidate = (storage_object_t *) object;



	LOG_ENTRY();

	/* The spare must be a disk, segment, or region. */
	if ((object_tag != DISK_TAG) &&
	    (object_tag != SEGMENT_TAG) &&
	    (object_tag != REGION_TAG)) {
		LOG_ERROR("Object %s is not a disk, segment, nor region.\n", mirror_candidate->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	/* The region cannot be a spare of itself. */
/*	if (mirror_candidate == (storage_object_t *)object->region) {
		LOG_ERROR("Region %s cannot be a mirror object for itself.\n", mirror_candidate->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
*/
	LOG_EXIT_INT(0);
	return 0;
}
#endif

/* aix_setup_evms_plugin
 *
 *	This function gets called shortly after the plugin is loaded by the
 *	Engine. It performs all tasks that are necessary before the initial
 *	discovery pass.
 */
static int aix_setup_evms_plugin(engine_functions_t     * functions)
{
	EngFncs = functions;
	my_plugin_record = &AIX_Plugin;


	LOG_ENTRY();

	// Allocate the global container list.
	aix_container_list = CreateList();
	if ( ! aix_container_list ) {
		LOG_CRITICAL("Error creating AIX container list.\n");
		return EINVAL;
	}

	LOG_EXIT_INT(0);

	return 0;
}


/* aix_cleanup_evms_plugin
 *
 *	Memory cleanup before closing the engine. At cleanup-time, we can only
 *	delete private data. All of the containers and objects must be left
 *	around, and the engine will free those before it exits.
 */
static void aix_cleanup_evms_plugin( void )
{
	storage_container_t     * container;
	TAG                     tag;

	LOG_ENTRY();

	GoToStartOfList(aix_container_list);
	while ( ! BlindExtractObject(aix_container_list, &tag, NULL, (void**)&container) ) {
		aix_deallocate_container(container, FALSE);
	}

	DestroyList(&aix_container_list, FALSE);

	LOG_EXIT_VOID();
}
/* Function:  aix_get_plugin_functions
 */
static int aix_get_plugin_functions(storage_object_t        *  object,
				    function_info_array_t   ** functions) {

	int rc = 0;
	function_info_array_t * fia;
	int function_count;
	uint count;
	//dlist_t available_objects = NULL;
	aix_region_data_t * r_data;
	//storage_object_t * tmp_object;

	LOG_ENTRY();


	/*
	 * If object is NULL, that means the user is asking for plug-in
	 * functions on the plug-in.  We don't have any plug-in functions that
	 * are global for the plug-in.
	 */

	if (object == NULL){
		LOG_EXIT_INT(ENOSYS);
		return ENOSYS;
	}

	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;

	r_data = object->private_data;

	fia = EngFncs->engine_alloc(sizeof(function_info_array_t) + (sizeof(function_info_t) * (AIX_MIRROR_FUNCTION_COUNT)));

	if (fia != NULL) {
		function_count = 0;

		/*
		 * Our functions are only allowed if the Engine is opened for
		 * writing.
		 */
		if (EngFncs->get_engine_mode() == ENGINE_READWRITE) {

			if (r_data->lv->mirror == AIX_DEFAULT_MIRRORING) {

				rc = aix_get_available_objects(object, &count);

				if ((count > 0) && (rc == 0)) {

					fia->info[function_count].function = AIX_MIRROR_FUNCTION_ADD_MIRROR;
					SET_STRING(fia->info[function_count].name, "addmirror");
					SET_STRING(fia->info[function_count].title, "Add mirror objects");
					SET_STRING(fia->info[function_count].verb, "Add");
					SET_STRING(fia->info[function_count].help, "Use this function to add an object as a mirror object for this LV.");

					function_count++;
				}

				/*
				 * An error from PruneList() or GetListSize()
				 * doesn't count.  It doesn't fail this
				 * function; it only means Add Mirror is not
				 * Available.
				 */
			}

			if (r_data->lv->mirror > AIX_DEFAULT_MIRRORING) {

				fia->info[function_count].function = AIX_MIRROR_FUNCTION_SPLIT_MIRROR;
				SET_STRING(fia->info[function_count].name, "splitmirror");
				SET_STRING(fia->info[function_count].title, "Split mirror object");
				SET_STRING(fia->info[function_count].verb, "Split");
				SET_STRING(fia->info[function_count].help, "Use this function to split objects mirrored to each other");

				function_count++;
			}

			rc = 0;
		}

		fia->count = function_count;
		*functions = fia;

	} else {
		LOG_CRITICAL("Error allocating memory for an action info array.\n");
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

/* Function:  aix_plugin_function
 */
static int aix_plugin_function(storage_object_t * object,
			       task_action_t      action,
			       dlist_t            objects,
			       option_array_t   * options)
{

	int rc = 0;
	aix_region_data_t    * r_data = object->private_data;
	storage_object_t     * mirror_object = NULL;
	storage_object_t     * mirror_object2 = NULL;
	storage_object_t     * tmp_object = NULL;
	storage_container_t  * container = object->producing_container;
	uint count = 0;
	int AllOnOnePV = TRUE;
	int i;

	LOG_ENTRY();

	LOG_DEBUG("object->name:%s for mirror add\n", object->name);

	if ((action < EVMS_Task_Plugin_Function) ||
	    (action > EVMS_Task_Plugin_Function + AIX_MIRROR_FUNCTION_COUNT)) {
		LOG_ERROR("Action code 0x%x is out of range.\n", action);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	switch (action) {
	case AIX_MIRROR_FUNCTION_ADD_MIRROR:


		if (objects == NULL) {
			LOG_ERROR("Must specify at least one mirror object to be added.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		rc = GetListSize(objects, &count);

		if ((rc == DLIST_SUCCESS) && (count > AIX_MAX_MIRRORS-1)) {
			LOG_ERROR("Can only specify up to %d object(s) to add as mirrors.\n", AIX_MAX_MIRRORS-1);
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		if (rc != 0) {
			LOG_EXIT_INT(rc);
			return rc;
		}
		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:


		if (objects == NULL) {
			LOG_ERROR("Must specify at least one mirror object to be split.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		break;


	default:
		LOG_ERROR("0x%x is not a valid action code.\n", action);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
		break;
	}


	switch (action) {
	case AIX_MIRROR_FUNCTION_ADD_MIRROR:
		// Is this object contained on one PV, if so, we'd really rather not put the mirror
		// on the same PV unless the user is queried and wants to continue.

		aix_check_physical_layout(object, &AllOnOnePV);

		i = 1;

		DLIST_FOR_EACH(tmp_object, objects, rc){
			switch (i) {
			case 1:
				mirror_object = tmp_object;
				break;
			case 2:
				mirror_object2 = tmp_object;
				break;
			default:
				break;
			}
			i++;
		}

		rc = aix_add_mirrors_to_object(object, mirror_object, mirror_object2, AllOnOnePV, container);

		if (rc) {
			MESSAGE("Error allocating new partitions to the region.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:
		rc = 0;

		rc = aix_split_mirror_region(object, options);

		if (rc) {
			MESSAGE("Error spliting the region.\n");
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}

		break;

	default:
		/*
		 * Shouldn't get here if the validiation
		 * code above did its job.
		 */
		LOG_WARNING("Action code 0x%x slipped past validation.\n", action);
		rc = EINVAL;
		break;
	}

	if (rc == DLIST_END_OF_LIST) {
		rc = DLIST_SUCCESS;
	}

	if (rc == 0) {
		aix_update_container_timestamp(container);
		r_data->flags |= SCFLAG_DIRTY;
		object->flags |= SCFLAG_DIRTY;
		object->producing_container->flags |= SCFLAG_DIRTY;
	}

	LOG_EXIT_INT(rc);
	return rc;
}
/****** Region Checking Functions ******/

/* All of the following aix_can_ functions return 0 if they are able to
 * perform the specified action, or non-zero if they cannot.
 */



/* aix_can_delete
 *
 *	Can we remove the specified AIX region, and consolidate the space with
 *	the corresponding freespace region? The region must belong to AIX, and
 *	must not be freespace.
 */
static int aix_can_delete( storage_object_t * region )
{
	LOG_ENTRY();

	// The region must belong to AIX.
	if ( region->plugin != my_plugin_record ) {
		LOG_ERROR("Region %s is not owned by AIX.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Must be a DATA region.
	if ( region->data_type != DATA_TYPE ) {
		LOG_DETAILS("Cannot delete freespace region %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_can_expand
 *
 *	Can this region be expanded, using freespace from its container? If so,
 *	add the region to the expansion_points output list.
 */
static int aix_can_expand(storage_object_t      * region,
			  u_int64_t               * expand_limit,
			  dlist_t                 expansion_points )
{
	aix_container_data_t  * c_data = region->producing_container->private_data;
	aix_region_data_t     * r_data = region->private_data;
	expand_object_info_t  * expand_point;
	void                  * handle;
	int                     rc = 0;
	u_int64_t               temp = 0;


	LOG_ENTRY();

	// Don't expand the freespace region

	if ( region->data_type != DATA_TYPE ) {
		LOG_DEBUG("Cannot expand freespace region %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if (r_data->lv->mirror != AIX_DEFAULT_MIRRORING) {
		LOG_DEBUG("Expanding mirror region currently not implemented %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Must be space available in the group.
	if (c_data->freespace->size == 0 ) {
		LOG_DEBUG("No freespace left in container %s.\n", region->producing_container->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Create a new expansion info object.
	expand_point = EngFncs->engine_alloc(sizeof(expand_object_info_t));
	if ( ! expand_point ) {
		LOG_CRITICAL("Memory error creating expansion info object.\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	temp = region->producing_container->size;
	LOG_DEBUG("producing_container_size = %Ld\n",temp);

	expand_point->object = region;
	expand_point->max_expand_size = *(expand_limit);

	if ( *expand_limit < region->size ) {
		expand_point->max_expand_size = round_down(*expand_limit, PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size));
	}
	else {
		expand_point->max_expand_size = c_data->freespace->size;
	}

	// Add to the expansion points list.
	rc = InsertObject(expansion_points,
			  expand_point,
			  EXPAND_OBJECT_TAG, NULL,
			  AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding region %s to expansion points list.\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_can_expand_by
 *
 *	AIX does not allow its child objects to expand.
 */
static int aix_can_expand_by(   storage_object_t        * child_object,
				u_int64_t               * size )
{

	LOG_ENTRY();

	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}


/* aix_can_shrink
 */
static int aix_can_shrink(storage_object_t      * region,
			  u_int64_t               * shrink_limit,
			  dlist_t                 shrink_points )
{
	aix_region_data_t       * r_data = region->private_data;
	aix_container_data_t    * c_data = region->producing_container->private_data;
	shrink_object_info_t    * shrink_point;
	void                            * handle;
	int                     rc;

	LOG_ENTRY();

	LOG_EXIT_INT(EINVAL); // MJR ... not working yet
	return EINVAL;

	// Don't shrink the freespace region

	if ( region->data_type != DATA_TYPE ) {
		LOG_DEBUG("Cannot shrink freespace region %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	if (r_data->lv->mirror != AIX_DEFAULT_MIRRORING) {
		LOG_DEBUG("Shrinking mirror region currently not implemented %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Create a new shrink info object.
	shrink_point = EngFncs->engine_alloc(sizeof(shrink_object_info_t));
	if ( ! shrink_point ) {
		LOG_CRITICAL("Memory error creating shrink info object.\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	shrink_point->object = region;
	shrink_point->max_shrink_size = (r_data->lv->num_lps - 1) * PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	if ( *shrink_limit < shrink_point->max_shrink_size ) {
		shrink_point->max_shrink_size = round_down(*shrink_limit, PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size));
	}

	// Add to the shrink points list.
	rc = InsertObject(shrink_points,
			  shrink_point,
			  SHRINK_OBJECT_TAG, NULL,
			  AppendToList, TRUE, &handle);
	if (rc) {
		LOG_SERIOUS("Error adding region %s to shrink points list.\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}


	LOG_EXIT_INT(0);
	return 0;
}


/* aix_can_shrink_by
 */
static int aix_can_shrink_by(   storage_object_t        * region,
				u_int64_t               * size )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(EINVAL);
	return EINVAL;
}


/* aix_can_set_volume
 *
 *	Any non-freespace AIX region can be converted into an EVMS
 *	compatibility volume.
 */
static int aix_can_set_volume(  storage_object_t        * region,
				boolean                 flag )
{
	LOG_ENTRY();

	// The region must belong to AIX.
	if ( region->plugin != my_plugin_record ) {
		LOG_ERROR("Region %s does not belong to AIX.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if ( (region->flags & SOFLAG_ACTIVE)==0 &&
	     (region->flags & SOFLAG_NEEDS_ACTIVATE)==0) {
		LOG_DETAILS("cant activate a volume on a region that we cant activate\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Freespace regions cannot become volumes.
	if ( region->data_type != DATA_TYPE ) {
		LOG_DETAILS("Cannot turn freespace region %s into a volume.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_discover
 *
 *	Examine all objects and find AIX PVs. Assemble containers
 *	and export all AIX LVs as EVMS regions.
 *
 *	All newly created regions must be added to the output list, and all
 *	objects from the input list must either be claimed or moved to the
 *	output list.
 */
static int aix_discover(dlist_t objects,
			dlist_t regions,
			boolean final_call )
{
	int rc;

	LOG_ENTRY();

	aix_discover_containers(objects, regions);

//	aix_check_containers();

	aix_read_vgdas();

	aix_discover_regions();

//	aix_check_regions();

	rc = aix_export_regions(regions);

	LOG_DEBUG("AIX Discovery Complete \n");

	LOG_EXIT_INT(rc);
	return rc;
}



/****** Region Functions ******/



/* aix_create
 *
 *	Create a new AIX region using the specified freespace.
 */
static int aix_create(  dlist_t         freespace_region_list,
			option_array_t  * options,
			dlist_t         new_region_list )
{
	storage_container_t     * container;
	storage_object_t        * free_region;
	storage_object_t        * new_region;
	aix_container_data_t    * c_data;
	lv_entries              * lv;
	aix_region_options_t    region_opts;
	int                     rc;

	LOG_ENTRY();

	// Get the selected freespace region.
	free_region = aix_get_freespace_region(freespace_region_list);
	if ( ! free_region ) {
		MESSAGE("Error getting freespace region for region creation.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	container = free_region->producing_container;
	c_data = container->private_data;
	LOG_DETAILS("Creating new region in container %s.\n", container->name);

	memset(&region_opts, 0, sizeof(aix_region_options_t));

	// Parse and verify the option array.
	rc = aix_create_region_parse_options(options, &region_opts);
	if (rc) {
		MESSAGE("Error parsing and verifying options.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Make sure there is enough space for the desired allocation scheme.
	rc = aix_check_available_partitions(container, &region_opts);
	if (rc) {
		MESSAGE("Not enough space available for desired allocation scheme.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Initialize a new LV entry.
	lv = aix_initialize_new_lv_entry(container, &region_opts);
	if ( ! lv ) {
		MESSAGE("Error initializing the LV metadata.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Allocate and initialize the new region.
	new_region = aix_allocate_region(lv, region_opts.name, c_data->vgda->vg_head->pp_size);
	if ( ! new_region ) {
		aix_clear_lv_entry(lv);
		MESSAGE("Error allocating new region.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Allocate PPs to the new region.
	rc = aix_allocate_partitions_to_region(new_region, container, &region_opts, AIX_CREATE_REGION);
	if (rc) {
		aix_deallocate_region(new_region, TRUE);
		aix_clear_lv_entry(lv);
		MESSAGE("Error allocating partitions to the new region.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Add the new region to the container and update all info.
	aix_add_object_to_list(new_region, container->objects_produced, AppendToList);
	new_region->producing_container = container;
	c_data->regions[lv->lvname] = new_region;
	c_data->vgda->vg_head->numlvs++;

	// Update the freespace region and rebuild the LP maps.
	rc = aix_update_freespace_region(container);
	if (rc) {
// Ack!!! We've already allocated PPs at this point, and
// added the new region to the container. Lots to fix up!
		MESSAGE("Error updating freespace for container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	aix_build_lp_maps(container, NULL, NULL, FALSE);

	// Update the container timestamps, mark it dirty,
	// and return the new region.
	aix_update_container_timestamp(container);
	container->flags |= SCFLAG_DIRTY;
	aix_add_object_to_list(new_region, new_region_list, AppendToList);

	// mark region dirty for commit and needs DM activate
	new_region->flags |= SOFLAG_NEEDS_ACTIVATE;

	LOG_DEBUG("Successfully created region %s.\n", new_region->name);
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_delete
 *
 *	Remove the specified region and consolidate all of its space into
 *	the corresponding freespace region.
 */
static int aix_delete(  storage_object_t        * region,
			dlist_t                 children )
{
	storage_container_t     * container;
	aix_container_data_t    * c_data;
	aix_region_data_t       * r_data;
	int             rc;

	LOG_ENTRY();
	LOG_DEBUG("Attempting to delete region %s.\n", region->name);

	// Make sure this region can be removed.
	rc = aix_can_delete(region);
	if (rc) {
		MESSAGE("Not allowed to delete region %s.\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	container = region->producing_container;
	r_data = region->private_data;
	c_data = container->private_data;

	// Remove all LPs from the region.
	aix_deallocate_partitions_from_region(region, 0);

	// Remove this region from its container.
	aix_remove_object_from_list(region, container->objects_produced);
	if ( c_data->regions[r_data->lv->lvname] == region ) {
		c_data->regions[r_data->lv->lvname] = NULL;
	}
	c_data->vgda->vg_head->numlvs--;

	// Delete the region from memory.
	aix_clear_lv_entry(r_data->lv);
	aix_deallocate_region(region, TRUE);

	// Update the freespace.
	aix_update_freespace_region(container);
	aix_build_lp_maps(container, NULL, NULL, FALSE);

	// Update the container's timestamp and mark dirty.
	aix_update_container_timestamp(container);
	container->flags |= SCFLAG_DIRTY;

	LOG_DEBUG("Successfully deleted region.\n");
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_expand
 */
static int aix_expand(  storage_object_t        * region,
			storage_object_t        * expand_object,
			dlist_t                 input_objects,
			option_array_t          * options )
{

	storage_object_t        * free_region;
	storage_container_t     * container;
	aix_region_options_t      region_opts;
	sector_count_t                expand_delta;
	aix_container_data_t    * c_data;
	int                               rc = 0;

	LOG_ENTRY();

	// If region and expand_object are not the same, someone is trying to
	// expand an object below LVM. Not allowed.
	if ( region != expand_object ) {
		LOG_ERROR("Cannot expand object %s below AIX region %s.\n", expand_object->name, region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	// Get the selected freespace from the input objects list, and make
	// sure it belongs to the target object's group.
	free_region = aix_get_freespace_region(input_objects);
	if (! free_region) {
		MESSAGE("Error getting freespace region for region expand. \n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	container = free_region->producing_container;
	c_data    = container->private_data;

	if (region->consuming_container != free_region->consuming_container) {
		LOG_ERROR("Target region and selected freespace are not in same group.\n");
		LOG_ERROR("Target of expand is %s.\n", region->name);
		LOG_ERROR("Selected freespace is %s.\n", c_data->freespace->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if ( region->flags & AIX_REGION_INCOMPLETE ) {
		LOG_ERROR("Region %s is currently missing PVs. Cannot expand.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	rc = aix_expand_region_parse_options(options, region, &region_opts);
	if (rc) {
		LOG_ERROR("Error parsing and verifying expand options\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Call the engine's can_expand_by() to make sure all plugins
	// and FSIMS above AIX LVM approve of the expand.
	expand_delta = region_opts.add_size;

	rc = EngFncs->can_expand_by(region, &expand_delta);
	if (rc) {
		LOG_ERROR("A parent of AIX LVM has disallowed the expand of region %s\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Make sure there is enough space for the desired allocation scheme.
	rc = aix_check_available_partitions(container, &region_opts);
	if (rc) {
		MESSAGE("Not enough space available for desired allocation scheme.\n");
		MESSAGE("Unable to create new region in container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}


	// Allocate new PPs to the region.
	rc = aix_expand_allocate_partitions_to_region(region, container, &region_opts, AIX_EXPAND_REGION);
	if (rc) {
		MESSAGE("Error allocating new partitions to the region.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Update the freespace region and rebuild the LP maps.
	rc = aix_update_freespace_region(container);
	if (rc) {

		MESSAGE("Error updating freespace for container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	aix_build_lp_maps(container, NULL, NULL, TRUE);

	// Update the container timestamps, mark it dirty,
	// and return the new region.
	aix_update_container_timestamp(container);
	container->flags |= SCFLAG_DIRTY;

	AIX_dump_expand_region(region,__LINE__);

	// mark region dirty for commit and needs DM activate
	region->flags |= SOFLAG_NEEDS_ACTIVATE;

	LOG_EXIT_INT(rc);

	return rc;
}


/* aix_shrink
 */
static int aix_shrink(  storage_object_t        * region,
			storage_object_t        * shrink_object,
			dlist_t                 input_objects,
			option_array_t          * options )
{
	aix_container_data_t * c_data = region->producing_container->private_data;
	aix_region_data_t    * r_data = region->private_data;
	sector_count_t          remove_extents = 0;
	sector_count_t          shrink_delta;
	aix_region_options_t      region_opts;
	int                     rc = 0;

	LOG_ENTRY();

	// If region and shrink_object are not the same, someone is trying to
	// shrink an object below AIX. Not allowed.
	if ( region != shrink_object ) {
		LOG_ERROR("Cannot shrink object %s below AIX region %s.\n", shrink_object->name, region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}


	// Don't shrink the freespace region, a volume involved in
	// snapshotting, or an incomplete volume.
	if ( region->data_type != DATA_TYPE ) {
		LOG_ERROR("Cannot shrink freespace region %s.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if ( region->flags & AIX_REGION_INCOMPLETE ) {
		LOG_ERROR("Region %s is currently missing PVs. Cannot shrink.\n", region->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Parse the options for the shrink command.
	rc = aix_shrink_region_parse_options(options, region, &region_opts);
	if (rc) {
		LOG_ERROR("Error parsing and verifying shrink options\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	remove_extents = region_opts.num_lps;

	// Call the engine's can_shrink_by() to make sure all plugins
	// and FSIMS above LVM approve of the shrink.
	shrink_delta = remove_extents * PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	rc = EngFncs->can_shrink_by(region, &shrink_delta);
	if (rc) {
		LOG_ERROR("A parent of LVM has disallowed the shrink of region %s\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Deallocate the necessary extents from the volume.
	rc = aix_deallocate_partitions_from_region(region, (r_data->lv->num_lps -remove_extents));
	if (rc) {
		LOG_ERROR("Unable to deallocate enough PEs to shrink region %s\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Update the freespace region and rebuild the LP maps.
	rc = aix_update_freespace_region(region->producing_container);
	if (rc) {

		MESSAGE("Error updating freespace for container %s.\n", region->producing_container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	aix_build_lp_maps(region->producing_container, NULL, NULL, FALSE);

	// Update the container timestamps, mark it dirty,
	// and return the new region.
	aix_update_container_timestamp(region->producing_container);
	region->producing_container->flags |= SCFLAG_DIRTY;
	region->flags |= SCFLAG_DIRTY;

	// Update the freespace and rebuild the LE maps.
	rc = aix_update_freespace_region(region->producing_container);
	if (rc) {
		LOG_ERROR("Error updating freespace for container %s\n", region->producing_container->name);
	}

	// mark region dirty for commit and needs DM activate
	region->flags |= SOFLAG_NEEDS_ACTIVATE;

	LOG_EXIT_INT(rc);

	return rc;
}


/* aix_set_volume
 */
static void aix_set_volume(     storage_object_t        * region,
				boolean                 flag )
{
	LOG_ENTRY();
	LOG_DEBUG("AIX set volume\n");
	LOG_EXIT_VOID();
}



/* aix_add_sectors_to_kill_list
 *
 *	The kill_sectors list contains a list of sectors that need to be zeroed
 *	during the next commit. This function is very similar to read/write.
 */
static int aix_add_sectors_to_kill_list( storage_object_t       * region,
					 lsn_t                   lsn,
					 sector_count_t          count )
{

	aix_region_data_t           * r_data = region->private_data;
	aix_container_data_t    * c_data = region->producing_container->private_data;

	lsn_t                   new_lsn = 0;
	sector_count_t          new_count;
	int                     rc = 0;
	storage_object_t * pv_entry = NULL;

	LOG_ENTRY();

	// Region cannot be a snapshot, a snapshot original, or
	// freespace. Don't kill past the end of the region either.
	if ( region->data_type != DATA_TYPE ||
	     lsn + count > r_data->lv->num_lps * GET_PHYSICAL_PART_SIZE(c_data->vgda->vg_head->pp_size) ) {
		LOG_EXIT_INT(EIO);
		return EIO;
	}

	// I/Os needs to deal with the possibility that a request can
	// come in that spans PEs or stripes. This is possible because
	// there is no limit on count. To fix this, we loop through
	// lvm_remap_sector and KILL_SECTORS until count reaches zero.
	// Also make sure to check for missing LE entries.
	while ( count ) {
		AIX_remap_sector(region, lsn, count, &new_lsn, &new_count, &pv_entry);
		if ( pv_entry && new_lsn ) {
			rc = KILL_SECTORS(pv_entry, new_lsn, new_count);
		}
		count   -= new_count;
		lsn     += new_count;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_commit_changes
 */
static int aix_commit_changes(  storage_object_t        * region,
				unsigned int            phase )
{
	LOG_ENTRY();
	LOG_DEBUG("AIX commit changes\n");
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_get_option_count
 *
 *	Determine the type of Task that is being performed, and return
 *	the number of options that are available for that Task.
 */
static int aix_get_option_count( task_context_t * context )
{
	int count;

	LOG_ENTRY();

	switch (context->action) {
	
	case EVMS_Task_Create:
		count = AIX_OPTION_REGION_NUM_LPS_INDEX + 1;
		break;

	case EVMS_Task_Create_Container:
		count = AIX_OPTION_CONTAINER_PP_SIZE_INDEX + 1;
		break;

	case EVMS_Task_Expand_Container:
		count = 0;
		break;

	case EVMS_Task_Expand:
		count = AIX_OPTION_REGION_NUM_LPS_INDEX + 1;
		break;

	case EVMS_Task_Shrink:
		count = AIX_OPTION_REGION_NUM_LPS_INDEX;
		break;


	case AIX_MIRROR_FUNCTION_ADD_MIRROR:
		count = 0;
		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:
		count = 0;
		break;

	default:
		count = -1;
		break;
	}

	LOG_EXIT_INT(count);
	return count;
}


/* aix_init_task
 *
 *	Determine the type of Task that is being performed, and set up the
 *	context structure with the appropriate initial values.
 */
static int aix_init_task( task_context_t * context )
{
	int rc = 0;

	aix_container_data_t * c_data;

	LOG_ENTRY();

	switch(context->action) {
	
	case EVMS_Task_Create:
		rc = aix_create_region_allocate_option_descriptor(context->option_descriptors);
		if (!rc) {
			rc = aix_create_region_get_acceptable(context->acceptable_objects);
			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
		}
		break;

	case EVMS_Task_Create_Container:
		rc = aix_create_container_allocate_option_descriptor(context->option_descriptors);
		if (!rc) {
			rc = aix_create_container_get_acceptable(context->acceptable_objects);
			context->min_selected_objects = 1;
			context->max_selected_objects = AIX_MAX_PVS;
		}
		break;


	case EVMS_Task_Expand_Container:
		// If the container already has the max number of PVs, then no
		// segments are acceptable.
		c_data = context->container->private_data;

		if ( c_data->vgda->vg_head->numpvs >= AIX_MAX_PVS ) {
			LOG_ERROR("Container %s already has maximum number of PVs (%d).\n", context->container->name, c_data->vgda->vg_head->numpvs);
			rc = EINVAL;
			break;
		}

		rc = aix_expand_container_allocate_option_descriptor(context->option_descriptors);
		if (!rc) {
			rc = aix_expand_container_get_acceptable(context->container, context->acceptable_objects);
			context->min_selected_objects = 1;
			context->max_selected_objects = AIX_MAX_PVS - c_data->vgda->vg_head->numpvs;
		}
		break;

	case EVMS_Task_Expand:

		// Don't allow expanding the freespace region
		if ( context->object->data_type == FREE_SPACE_TYPE ) {
			LOG_ERROR("Cannot expand freespace region %s\n", context->object->name);
			rc = EINVAL;
			break;
		}

		rc = aix_expand_region_allocate_option_descriptor(context->option_descriptors);
		if (!rc) {
			rc = aix_expand_region_get_acceptable(context->object, context->acceptable_objects);
			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
		}
		break;

	case EVMS_Task_Shrink:
		// Don't allow shrinking the freespace region.
		if ( context->object->data_type == FREE_SPACE_TYPE ) {
			LOG_WARNING("Cannot shrink freespace region %s\n", context->object->name);
			rc = EINVAL;
			break;
		}

		rc = aix_shrink_region_allocate_option_descriptor(context->option_descriptors);
		if (!rc) {
			rc = aix_shrink_region_get_acceptable(context->object, context->acceptable_objects);
			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
		}
		break;

	case AIX_MIRROR_FUNCTION_ADD_MIRROR:
		{
			int answer = NO_MIRROR;
			char *choice_text[3] = {"No", "Continue", NULL};
			char *mirror_msg = "You are attempting to create a mirrored object in a Container with only one physical volume or partition.\
                               A disk failure will result in the loss of ALL copies of the data!";

			c_data = context->object->producing_container->private_data;

			LOG_DEBUG("Selected add mirror from context menu\n");

			if (c_data->vgda->vg_head->numpvs == 1) {
				EngFncs->user_message(my_plugin_record, &answer, choice_text, mirror_msg);
			}
			else {
				answer = MIRROR_CONTINUE;
			}

			if (answer == MIRROR_CONTINUE) {
				rc = aix_list_acceptable_mirrors(context->object, context->acceptable_objects);
				context->min_selected_objects = 1;
				context->max_selected_objects = 2;
			}

		}
		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:
		{
			rc = aix_split_region_allocate_option_descriptor(context->option_descriptors);
			rc = 0;
		}
		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_set_option
 *
 *	Determine the type of Task that is being performed. Then examine the
 *	desired option (using the index), and verify that the given value is
 *	appropriate. Reset the value if necessary and possible. Adjust other
 *	options as appropriate.
 */
static int aix_set_option(task_context_t        * context,
			  u_int32_t               index,
			  value_t                 * value,
			  task_effect_t           * effect )
{
	int rc;

	LOG_ENTRY();

	*effect = 0;

	switch (context->action) {
	
	case EVMS_Task_Create:
		rc = aix_create_region_set_option(context, index, value, effect);
		break;

	case EVMS_Task_Create_Container:
		rc = aix_create_container_set_option(context, index, value, effect);
		break;


	case EVMS_Task_Expand_Container:
		rc = aix_expand_container_set_option(context, index, value, effect);
		break;

	case EVMS_Task_Expand:
		rc = aix_expand_region_set_option(context, index, value, effect);
		break;

	case EVMS_Task_Shrink:
		rc = aix_shrink_region_set_option(context, index, value, effect);
		break;

	case AIX_MIRROR_FUNCTION_ADD_MIRROR:
		rc = 0;
		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:
		rc = aix_split_region_set_option(context, index, value, effect);
		rc = 0;
		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_auto_select_option
 *
 */
#if 0
static int aix_auto_select_option(      task_context_t  * context,
					u_int32_t       index,
					task_effect_t   * effect )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}
#endif


/* aix_set_objects
 *
 *	Determine the type of task, and then validate that the objects on the
 *	"selected" list are valid for that task. If so, adjust the option
 *	descriptor as appropriate.
 */
static int aix_set_objects(     task_context_t  * context,
				dlist_t         declined_objects,
				task_effect_t   * effect )
{
	int rc;

	LOG_ENTRY();

	*effect = 0;

	switch (context->action) {
	
	case EVMS_Task_Create:
		rc = aix_create_region_set_objects(context, declined_objects, effect);
		break;

	case EVMS_Task_Create_Container:
		rc = aix_create_container_set_objects(context, declined_objects, effect);
		break;

	case EVMS_Task_Expand_Container:
		rc = aix_expand_container_set_objects(context, declined_objects, effect);
		break;

	case EVMS_Task_Expand:
		rc = aix_expand_region_set_objects(context, declined_objects, effect);
		break;

	case EVMS_Task_Shrink:
		rc = aix_shrink_region_set_objects(context, declined_objects, effect);
		break;

	case AIX_MIRROR_FUNCTION_ADD_MIRROR:
		rc = 0;
		break;

	case AIX_MIRROR_FUNCTION_SPLIT_MIRROR:
		rc = 0;
		break;


	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);

	return(rc);
}


/* aix_get_info
 *
 *	Return AIX-specific information about the specified region. If the
 *	name field is set, only return the "extra" information pertaining
 *	to that name.
 */
static int aix_get_info(        storage_object_t        * region,
				char                    * name,
				extended_info_array_t   ** info_array )
{
	int                     rc = 0;

	LOG_ENTRY();

	if ( ! name ) {
		// Default case. Return all basic info about the region.
		rc = aix_get_volume_info(region, info_array);
	}
	else if ( ! strcmp(name, "Extents") ) {
		// "Extra" information about LE-to-PE mapping.
		rc = aix_get_volume_extent_info(region, info_array);
	}
	else {
		LOG_ERROR("Currently No support for extra region information about \"%s\"\n", name);
		rc = EINVAL;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_set_info
 */
static int aix_set_info(        storage_object_t        * region,
				option_array_t          * options )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}


/* aix_get_plugin_info
 *
 *	Return information about the AIX plugin. There is no "extra"
 *	information about AIX, so "name" should always be NULL.
 */
static int aix_get_plugin_info( char                    * name,
				extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
	char                    buffer[50] = {0};
	int                     i = 0;

	LOG_ENTRY();

	if ( ! name ) {
		// Get memory for the info array
		info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*6);
		if ( ! info ) {
			LOG_CRITICAL("Memory error creating info array\n");
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		// Short Name
		SET_STRING(info->info[i].name, "ShortName");
		SET_STRING(info->info[i].title, "Short Name");
		SET_STRING(info->info[i].desc, "A short name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, my_plugin_record->short_name);
		i++;

		// Long Name
		SET_STRING(info->info[i].name, "LongName");
		SET_STRING(info->info[i].title, "Long Name");
		SET_STRING(info->info[i].desc, "A long name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, my_plugin_record->long_name);
		i++;

		// Plugin Type
		SET_STRING(info->info[i].name, "Type");
		SET_STRING(info->info[i].title, "Plugin Type");
		SET_STRING(info->info[i].desc, "There are various types of plugins; each responsible for some kind of storage object.");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, "Region Manager");
		i++;

		// Plugin Version
		SET_STRING(info->info[i].name, "Version");
		SET_STRING(info->info[i].title, "Plugin Version");
		SET_STRING(info->info[i].desc, "This is the version number of the plugin.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Engine Services Version
		SET_STRING(info->info[i].name, "Required_Engine_Version");
		SET_STRING(info->info[i].title, "Required Engine Services Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine services that this plug-in requires.  It will not run on older versions of the Engine services.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", my_plugin_record->required_engine_api_version.major, my_plugin_record->required_engine_api_version.minor, my_plugin_record->required_engine_api_version.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Plug-in API Version
		SET_STRING(info->info[i].name, "Required_Plugin_Version");
		SET_STRING(info->info[i].title, "Required Plug-in API Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine plug-in API that this plug-in requires.  It will not run on older versions of the Engine plug-in API.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", my_plugin_record->required_plugin_api_version.plugin.major, my_plugin_record->required_plugin_api_version.plugin.minor, my_plugin_record->required_plugin_api_version.plugin.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Container API Version
		SET_STRING(info->info[i].name, "Required_Container_Version");
		SET_STRING(info->info[i].title, "Required Container API Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine container API that this plug-in requires.  It will not run on older versions of the Engine container API.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", my_plugin_record->required_container_api_version.major, my_plugin_record->required_container_api_version.minor, my_plugin_record->required_container_api_version.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;
		info->count = i;

		*info_array = info;
	}
	else {
		LOG_ERROR("No support for extra plugin information about \"%s\"\n", name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_read
 *
 *	Perform a logical-to-physical remapping, and send the read down to
 *	the next plugin.
 */
static int aix_read(    storage_object_t        * region,
			lsn_t                   lsn,
			sector_count_t          count,
			void                    * buffer )
{

	aix_region_data_t           * r_data = region->private_data;
	aix_container_data_t    * c_data = region->producing_container->private_data;
	lsn_t                   new_lsn = 0;
	sector_count_t          new_count;
	int                     rc = 0;
	storage_object_t * pv_entry = NULL;

	LOG_ENTRY();


	// No reads allowed on snapshot volumes, incomplete volumes, or
	// freespace volumes. Don't read past the end of the volume.
	if ( region->data_type != DATA_TYPE ||
	     lsn + count > r_data->lv->num_lps * GET_PHYSICAL_PART_SIZE(c_data->vgda->vg_head->pp_size) ) {
		memset(buffer, 0, sectors_to_bytes(count));
		LOG_EXIT_INT(0);
		return 0;
	}

	// I/Os needs to deal with requests that can span PEs
	// or stripes. This is possible because there is no
	// limit on count. To fix this, we loop through
	// lvm_remap_sector and READ until count reaches zero.
	// Also make sure to check for missing LE entries.
	while ( count ) {
		AIX_remap_sector(region, lsn, count, &new_lsn, &new_count, &pv_entry);
		if ( pv_entry && new_lsn ) {
			rc = READ(pv_entry, new_lsn, new_count, buffer);
		}
		else {
			memset(buffer, 0, sectors_to_bytes(new_count));
			rc = EIO;
		}
		count   -= new_count;
		lsn     += new_count;
		buffer  = (void*)(((unsigned long)buffer) + (unsigned long)(new_count << SECTOR_SIZE_BITS));
	}

	LOG_EXIT_INT(rc);
	return rc;

}


/* aix_write
 *
 *	Perform a logical-to-physical remapping, and send the write down to
 *	the next plugin.
 */
static int aix_write(   storage_object_t        * region,
			lsn_t                   lsn,
			sector_count_t          count,
			void                    * buffer )
{

	aix_region_data_t           * r_data = region->private_data;
	aix_container_data_t    * c_data = region->producing_container->private_data;
	lsn_t                   new_lsn;
	sector_count_t          new_count;
	int                     rc = 0;
	storage_object_t * pv_entry = NULL;

	LOG_ENTRY();

	// No writes allowed to any volumes involved in snapshotting, any
	// incomplete volumes, or any freespace volumes. Don't write past
	// the end of the volume.
	if ( region->data_type != DATA_TYPE ||
	     lsn + count > r_data->lv->num_lps * GET_PHYSICAL_PART_SIZE(c_data->vgda->vg_head->pp_size) ) {
		LOG_EXIT_INT(EIO);
		return EIO;
	}

	// I/Os needs to deal with requests that span PEs or
	// stripes. This is possible because there is no limit
	// on count. To fix this, we loop through
	// lvm_remap_sector and WRITE until count reaches zero.
	// Also check for missing LE entries.
	while ( count ) {
		AIX_remap_sector(region, lsn, count, &new_lsn, &new_count, &pv_entry);
		if ( pv_entry && new_lsn ) {
			rc = WRITE(pv_entry, new_lsn, new_count, buffer);
		}
		else {
			rc = EIO;
		}
		count   -= new_count;
		lsn     += new_count;
		buffer  = (void*)(((unsigned long)buffer) + (unsigned long)(new_count << SECTOR_SIZE_BITS));
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_direct_plugin_communication
 */
static int aix_direct_plugin_communication(     void    * thing,
						boolean target_kernel_plugin,
						void    * arg )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}



/****** Container Functions ******/



/* aix_can_create_container
 *
 *	Can we create a new container from this list of objects?
 */
static int aix_can_create_container( dlist_t objects )
{
	storage_object_t        * object;
	int                     rc;

	LOG_ENTRY();

	// Needs to be at least one object on the list. Don't create
	// an empty container.
	if ( ListEmpty(objects) ) {
		LOG_WARNING("No objects specified.\n");
		LOG_WARNING("Cannot create an empty container.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Check that each object can be included in a container.
	DLIST_FOR_EACH(object, objects, rc) {
		rc = aix_check_object_for_container_inclusion(object);
		if (rc) {
			LOG_WARNING("%s cannot be added to an AIX container.\n", object->name);
			LOG_EXIT_INT(rc);
			return rc;
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* : aix_can_delete_container
 *
 *	Can we remove the specified container?
 */
static int aix_can_delete_container( storage_container_t * container )
{
	aix_container_data_t * c_data = container->private_data;

	LOG_ENTRY();

	// Make sure the container belongs to AIX.
	if ( container->plugin != my_plugin_record ) {
		LOG_ERROR("Container %s does not belong to AIX.\n", container->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Check if there are any data regions in this container.
	if ( c_data->vgda->vg_head->numlvs > 0 ) {
		LOG_ERROR("Container %s is still exporting %d regions. Cannot be deleted.\n",
			  container->name, c_data->vgda->vg_head->numlvs);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_can_add_object
 */
static int aix_can_add_object(  storage_object_t        * object,
				storage_container_t     * container )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}


/* aix_can_remove_object
 */
static int aix_can_remove_object( storage_object_t * object )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}


/* aix_create_container
 *
 *	Create a new AIX container, using the specified objects.
 */
static int aix_create_container( dlist_t                        objects,
				 option_array_t                  * options,
				 storage_container_t             ** container )
{
	storage_object_t        * object;
	storage_container_t     * new_container;
	unsigned long           pp_size;	// in vsectors
	int                     rc;

	LOG_ENTRY();

	// Needs to be at least one object on the list. Don't create
	// an empty container.
	if ( ListEmpty(objects) ) {
		MESSAGE("No objects specified.\n");
		MESSAGE("Cannot create an empty container.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_DEBUG("Attempting to create new container with objects:\n");
	DLIST_FOR_EACH(object, objects, rc) {
		LOG_DEBUG("\t%s\n", object->name);
	}

	// Parse the options for container creation
	rc = aix_create_container_parse_options(options, &pp_size);
	if (rc) {
		MESSAGE("Error parsing and verifying options.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Perform necessary checks against all objects. Make sure the
	// object can be used in a container. Make sure the PP size is
	// valid for the object.
	DLIST_FOR_EACH(object, objects, rc) {
		rc = aix_check_object_for_container_inclusion(object);
		if (rc) {
			MESSAGE("%s cannot be added to this AIX container. "
				"Has a volume, parent objects, or is already in a container.\n", object->name);
			LOG_EXIT_INT(rc);
			return rc;
		}

		rc = aix_check_object_for_pp_size(object, &pp_size);
		if (rc) {
			MESSAGE("%s cannot be added to this AIX container. "
				"Object size is invalid for PP size.\n", object->name);
			LOG_EXIT_INT(rc);
			return rc;
		}
	}

	// Create the new container.
	new_container = aix_allocate_new_container(pp_size);
	if ( ! new_container ) {
		MESSAGE("Memory error allocating new container.\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Add each object to the container.
	DLIST_FOR_EACH(object, objects, rc) {
		// Create a PV header and PP map.
		rc = aix_create_pv_from_object(object, new_container);
		if (rc) {
			aix_deallocate_container(new_container, TRUE);
			MESSAGE("Error creating new PV for %s.\n", object->name);
			LOG_EXIT_INT(rc);
			return rc;
		}

		// Add this PV to the container.
		rc = aix_add_new_pv_to_container(object, new_container);
		if (rc) {
			aix_deallocate_container(new_container, TRUE);
			MESSAGE("Error adding %s to new container %s.\n", object->name, new_container->name);
			LOG_EXIT_INT(rc);
			return rc;
		}
	}

	// Update the freespace region for the container.
	rc = aix_update_freespace_region(new_container);
	if (rc) {
		aix_deallocate_container(new_container, TRUE);
		MESSAGE("Error updating freespace region in new container %s.\n", new_container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Rebuild the LP maps.
	aix_build_lp_maps(new_container, NULL, NULL, FALSE);

	// Update the container's timestamp
	aix_update_container_timestamp(new_container);

	// Mark the container dirty.
	new_container->flags |= SCFLAG_DIRTY;

	// Return the container to the engine.
	*container = new_container;

	LOG_DEBUG("Successfully created new container %s\n", new_container->name);
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_add_object
 *
 *	Add the specified object to the specified (existing) container. If
 *	container is NULL, add the object to the unassigned container. The
 *	object must have already been removed from its previous location
 *	before this function is called. This function is not used to move an
 *	object from one AIX container to another AIX container. The function
 *	aix_transfer_object will be used for that purpose.
 */
static int aix_add_object(      storage_object_t        * region,
				storage_container_t     * container,
				option_array_t          * options )
{

//	aix_region_data_t	    * r_data = region->private_data;
	aix_container_data_t    * c_data = container->private_data;

	unsigned long           pp_size = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	int                     rc;

	LOG_ENTRY();

	// Needs to be at least one object on the list. Don't create
	// an empty container.
	if ( !region ) {
		MESSAGE("No objects specified.\n");
		MESSAGE("Cannot add nothing to a container.\n");
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_DEBUG("Attempting to add to container with object:\t%s\n",region->name);

	// Parse the options for container creation
	rc = aix_expand_container_parse_options(options, &pp_size);
	if (rc) {
		MESSAGE("Error parsing and verifying options.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Perform necessary checks against all objects. Make sure the
	// object can be used in a container. Make sure the PP size is
	// valid for the object.
	rc = aix_check_object_for_container_inclusion(region);
	if (rc) {
		MESSAGE("%s cannot be added to this AIX container. Has a volume, "
			"parent objects, or is already in a container.\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}


	// Create a PV header and PP map.
	rc = aix_create_pv_from_object(region, container);
	if (rc) {
		MESSAGE("Error creating new PV for %s.\n", region->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Add this PV to the container.
	rc = aix_add_new_pv_to_container(region, container);
	if (rc) {
		MESSAGE("Error adding %s to new container %s.\n", region->name, container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Update the freespace region for the container.
	rc = aix_update_freespace_region(container);
	if (rc) {
		aix_deallocate_container(container, TRUE);
		MESSAGE("Error updating freespace region in new container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Rebuild the LP maps.
	aix_build_lp_maps(container, NULL, NULL, FALSE);

	// Update the container's timestamp
	aix_update_container_timestamp(container);

	// Mark the container dirty.
	container->flags |= SCFLAG_DIRTY;

	LOG_DEBUG("Successfully added object:%s to container %s\n", region->name, container->name);
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_transfer_object
 *
 *	Move the specified object from its current container to the specified
 *	container. Unlike add_object, this function is used to move objects
 *	around internally to AIX.
 */
static int aix_transfer_object( storage_object_t        * object,
				storage_container_t     * target_container,
				option_array_t          * options )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}


/* aix_remove_object
 *
 *	Remove the specified object from its current container. Make sure the
 * 	object does not contain any allocated regions. In addition to removing
 *	it from the container, we want to kill any metadata that might have
 *	been on the object.
 */
static int aix_remove_object( storage_object_t * object )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}


/* aix_delete_container
 *
 *	Remove the specified container.
 */
static int aix_delete_container(storage_container_t     * container,
				dlist_t                 objects_consumed )
{
	storage_object_t        * object;
	dlist_t                 temp_child_list;
	int                     rc;

	LOG_ENTRY();
	LOG_DEBUG("Attempting to delete container %s.\n", container->name);

	// Make sure this container can be deleted.
	rc = aix_can_delete_container(container);
	if (rc) {
		MESSAGE("Not allowed to delete container %s.\n", container->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Make a temporary copy of the container's child_list, since we
	// will be removing items from it during the FOR_EACH loop.
	temp_child_list = CreateList();
	if ( ! temp_child_list ) {
		MESSAGE("Unable to get copy of child list for container %s.\n", container->name);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
	CopyList(temp_child_list, container->objects_consumed, AppendToList);

	// Remove each PV from the container individually.
	DLIST_FOR_EACH(object, temp_child_list, rc) {
		LOG_DETAILS("Removing object %s from container %s.\n", object->name, container->name);

		// Remove this PV from the container and update the VG info.
		aix_remove_pv_from_container(object);

		// Deallocate the PV.
		aix_deallocate_pv(object);

		// Place this object on the output list.
		aix_add_object_to_list(object, objects_consumed, AppendToList);
	}

	// Free up the container's memory.
	aix_deallocate_container(container, TRUE);

	DestroyList(&temp_child_list, FALSE);
	LOG_DEBUG("Successfully deleted container.\n");
	LOG_EXIT_INT(0);
	return 0;
}


/* aix_commit_container_changes
 *
 *	Write all of the metadata for this container to the underlying objects.
 */
static int aix_commit_container_changes( storage_container_t    * container,
					 unsigned int            phase )
{
	storage_object_t        * object;
	aix_container_data_t * c_data = container->private_data;
	int                     rc, i, err = 0;

	LOG_ENTRY();

	// Does this container belong to AIX.
	if ( container->plugin != my_plugin_record ) {
		LOG_WARNING("Container %s does not belong to the AIX region manager.\n", container->name);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// Is this container dirty?
	if ( ! (container->flags & SCFLAG_DIRTY) ) {
		LOG_WARNING("Container %s is not dirty.\n", container->name);
		LOG_EXIT_INT(0);
		return 0;
	}

	// Two copies of metadata to write to each PV in the container. Stop
	// the commit on any error.
	DLIST_FOR_EACH(object, container->objects_consumed, rc) {
		switch (phase) {
		case 1:
			// Only one copy of the IPL and LVM.
			if (!err) err = aix_write_ipl(object);
			if (!err) err = aix_write_lvm(object);
		case 2:
			// Write the VGSA and VGDA.
			if (!err) err = aix_write_vgsa(object, phase-1);
			if (!err) err = aix_write_vg_header(object, phase-1);
			if (!err) err = aix_write_lv_array(object, phase-1);
			if (!err) err = aix_write_pvs(object, phase-1);
			if (!err) err = aix_write_lv_names(object, phase-1);
			if (!err) err = aix_write_vg_trailer(object, phase-1);
			break;
		default:
			break;
		}
	}

	// Mark the container and all regions within this container clean on the second pass.
	// If we got here w/o a problem then everything in this container is clean
	if (!err && phase == 2) {
		container->flags &= ~SCFLAG_DIRTY;
		LOG_DETAILS("Wrote container %s to disk.\n", container->name);

		for (i =0; i < AIX_MAX_LVS; i++) {
			if (c_data->regions[i]) {
				c_data->regions[i]->flags &= ~SCFLAG_DIRTY;
			}
		}
	}

	LOG_EXIT_INT(err);
	return err;
}


/* aix_get_container_info
 *
 *	Return an information array for the specified container. If the name
 *	field is set, only return the "extra" information pertaining to that
 *	name.
 */
static int aix_get_container_info(      storage_container_t     * container,
					char                    * name,
					extended_info_array_t   ** info_array )
{

	char                    * buf = NULL;
	int                     rc;

	LOG_ENTRY();

	if ( ! name ) {
		// Default case. Return all basic info about the container.
		rc = aix_get_group_info(container, info_array);
	}
	else if ( ! strcmp(name, "Current_PVs") ) {
		// Return a list of names of all PVs
		rc = aix_get_group_pv_list_info(container, info_array);
	}
	else if ( ! strcmp(name, "Current_LVs") ) {
		// Return a list of names of all LVs
		rc = aix_get_group_lv_list_info(container, info_array);
	}
	else if ( (buf = strstr(name, "PV")) == name ) {
		// PV-specific information. This is what we will
		// use for pvdisplay.
		rc = aix_get_pv_info(container, info_array);
	}
	else if ( (buf = strstr(name, "PEMapPV")) == name ) {
		// PV-specific extent information. This will be used
		// for pvdisplay -v
		rc = aix_get_pv_extent_info(container, info_array);
	}
	else {
		LOG_ERROR("No support for extra information about \"%s\"\n", name);
		rc = EINVAL;
	}


	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_set_container_info
 *
 *	Apply the settings of the options to the container.
 */
static int aix_set_container_info(      storage_container_t     * container,
					option_array_t          * options )
{
	LOG_ENTRY();
	LOG_DEBUG("Not yet implemented\n");
	LOG_EXIT_INT(ENOSYS);
	return ENOSYS;
}



/* Function tables for the AIX Region Manager */
static plugin_functions_t aix_functions = {
	setup_evms_plugin               : aix_setup_evms_plugin,
	cleanup_evms_plugin             : aix_cleanup_evms_plugin,
	can_delete                      : aix_can_delete,
	can_expand                      : aix_can_expand,
	can_expand_by                   : aix_can_expand_by,
	can_shrink                      : aix_can_shrink,
	can_shrink_by                   : aix_can_shrink_by,
	can_set_volume                  : aix_can_set_volume,
	discover                        : aix_discover,
	create                          : aix_create,
	delete                          : aix_delete,
	expand                          : aix_expand,
	shrink                          : aix_shrink,
	set_volume                      : aix_set_volume,
	add_sectors_to_kill_list 	: aix_add_sectors_to_kill_list,
	commit_changes  		: aix_commit_changes,
	get_option_count    		: aix_get_option_count,
	init_task                       : aix_init_task,
	set_option                      : aix_set_option,
	set_objects                     : aix_set_objects,
	get_info                        : aix_get_info,
	set_info                        : aix_set_info,
	get_plugin_info                 : aix_get_plugin_info,
	read                            : aix_read,
	write                           : aix_write,
	get_plugin_functions    	: aix_get_plugin_functions,
	plugin_function         	: aix_plugin_function,
	direct_plugin_communication     : aix_direct_plugin_communication,
	activate			: aix_activate_LV,
	deactivate			: aix_deactivate_LV
};


static container_functions_t aix_container_functions = {
	can_create_container            : aix_can_create_container,
	can_delete_container            : aix_can_delete_container,
	can_add_object                  : aix_can_add_object,
	can_remove_object               : aix_can_remove_object,
	create_container                : aix_create_container,
	add_object                      : aix_add_object,
	transfer_object                 : aix_transfer_object,
	remove_object                   : aix_remove_object,
	delete_container                : aix_delete_container,
	commit_container_changes        : aix_commit_container_changes,
	get_container_info              : aix_get_container_info,
	set_container_info              : aix_set_container_info
};


plugin_record_t AIX_Plugin = {
	id                              : SetPluginID( EVMS_OEM_IBM, EVMS_REGION_MANAGER, 0x03),
	version                         : {
		major                   : MAJOR_VERSION,
		minor                   : MINOR_VERSION,
		patchlevel              : PATCH_LEVEL
	},
	required_engine_api_version             : {
		major                   : 11,
		minor                   : 0,
		patchlevel              : 0
	},
	required_plugin_api_version             : {plugin       : {
			major                   : 10,
			minor                   : 0,
			patchlevel              : 0
		} },
	required_container_api_version          : {
		major                   : 8,
		minor                   : 0,
		patchlevel              : 0
	},
	short_name                      : "AixRegMgr",
	long_name                       : "AIX Region Manager",
	oem_name                        : "IBM",
	functions                       : {
		plugin                  : &aix_functions
	},
	container_functions             : &aix_container_functions
};

plugin_record_t * evms_plugin_records[] = {
	&AIX_Plugin,
	NULL
};

