/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libs390.so
 *
 *   File: helpers.c
 */
/*
 *  Special thanks to dasdfmt and fdasd code.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>

#include "vtoc.h"
#include "390segmgr.h"
#include "dm.h"
#include "helpers.h"


static unsigned char EBCtoASC[256] =
{
/* 0x00  NUL   SOH   STX   ETX  *SEL    HT  *RNL   DEL */
	0x00, 0x01, 0x02, 0x03, 0x07, 0x09, 0x07, 0x7F,
/* 0x08  -GE  -SPS  -RPT    VT    FF    CR    SO    SI */
	0x07, 0x07, 0x07, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
/* 0x10  DLE   DC1   DC2   DC3  -RES   -NL    BS  -POC
				-ENP  ->LF             */
	0x10, 0x11, 0x12, 0x13, 0x07, 0x0A, 0x08, 0x07,
/* 0x18  CAN    EM  -UBS  -CU1  -IFS  -IGS  -IRS  -ITB
						  -IUS */
	0x18, 0x19, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07,
/* 0x20  -DS  -SOS    FS  -WUS  -BYP    LF   ETB   ESC
				-INP                   */
	0x07, 0x07, 0x1C, 0x07, 0x07, 0x0A, 0x17, 0x1B,
/* 0x28  -SA  -SFE   -SM  -CSP  -MFA   ENQ   ACK   BEL
		     -SW                               */
	0x07, 0x07, 0x07, 0x07, 0x07, 0x05, 0x06, 0x07,
/* 0x30 ----  ----   SYN   -IR   -PP  -TRN  -NBS   EOT */
	0x07, 0x07, 0x16, 0x07, 0x07, 0x07, 0x07, 0x04,
/* 0x38 -SBS   -IT  -RFF  -CU3   DC4   NAK  ----   SUB */
	0x07, 0x07, 0x07, 0x07, 0x14, 0x15, 0x07, 0x1A,
/* 0x40   SP   RSP                         ----       */
	0x20, 0xFF, 0x83, 0x84, 0x85, 0xA0, 0x07, 0x86,
/* 0x48                      .     <     (     +     | */
	0x87, 0xA4, 0x9B, 0x2E, 0x3C, 0x28, 0x2B, 0x7C,
/* 0x50    &                                      ---- */
	0x26, 0x82, 0x88, 0x89, 0x8A, 0xA1, 0x8C, 0x07,
/* 0x58               !     $     *     )     ;       */
	0x8D, 0xE1, 0x21, 0x24, 0x2A, 0x29, 0x3B, 0xAA,
/* 0x60    -     /  ----       ----  ----  ----       */
	0x2D, 0x2F, 0x07, 0x8E, 0x07, 0x07, 0x07, 0x8F,
/* 0x68             ----     ,     %     _     >     ? */
	0x80, 0xA5, 0x07, 0x2C, 0x25, 0x5F, 0x3E, 0x3F,
/* 0x70  ---        ----  ----  ----  ----  ----  ---- */
	0x07, 0x90, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07,
/* 0x78    *     `     :     #     @     '     =     " */
	0x70, 0x60, 0x3A, 0x23, 0x40, 0x27, 0x3D, 0x22,
/* 0x80    *     a     b     c     d     e     f     g */
	0x07, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
/* 0x88    h     i              ----  ----  ----       */
	0x68, 0x69, 0xAE, 0xAF, 0x07, 0x07, 0x07, 0xF1,
/* 0x90         j     k     l     m     n     o     p */
	0xF8, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F, 0x70,
/* 0x98    q     r                    ----        ---- */
	0x71, 0x72, 0xA6, 0xA7, 0x91, 0x07, 0x92, 0x07,
/* 0xA0          ~     s     t     u     v     w     x */
	0xE6, 0x7E, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78,
/* 0xA8    y     z              ----  ----  ----  ---- */
	0x79, 0x7A, 0xAD, 0xAB, 0x07, 0x07, 0x07, 0x07,
/* 0xB0    ^                    ----       ----       */
	0x5E, 0x9C, 0x9D, 0xFA, 0x07, 0x07, 0x07, 0xAC,
/* 0xB8       ----     [     ]  ----  ----  ----  ---- */
	0xAB, 0x07, 0x5B, 0x5D, 0x07, 0x07, 0x07, 0x07,
/* 0xC0    {     A     B     C     D     E     F     G */
	0x7B, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
/* 0xC8    H     I  ----                         ---- */
	0x48, 0x49, 0x07, 0x93, 0x94, 0x95, 0xA2, 0x07,
/* 0xD0    }     J     K     L     M     N     O     P */
	0x7D, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50,
/* 0xD8    Q     R  ----                              */
	0x51, 0x52, 0x07, 0x96, 0x81, 0x97, 0xA3, 0x98,
/* 0xE0    \           S     T     U     V     W     X */
	0x5C, 0xF6, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
/* 0xE8    Y     Z        ----       ----  ----  ---- */
	0x59, 0x5A, 0xFD, 0x07, 0x99, 0x07, 0x07, 0x07,
/* 0xF0    0     1     2     3     4     5     6     7 */
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
/* 0xF8    8     9  ----  ----       ----  ----  ---- */
	0x38, 0x39, 0x07, 0x07, 0x9A, 0x07, 0x07, 0x07
};

static unsigned char ASCtoEBC[256] =
{
	/*00  NL    SH    SX    EX    ET    NQ    AK    BL */
	0x00, 0x01, 0x02, 0x03, 0x37, 0x2D, 0x2E, 0x2F,
	/*08  BS    HT    LF    VT    FF    CR    SO    SI */
	0x16, 0x05, 0x15, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	/*10  DL    D1    D2    D3    D4    NK    SN    EB */
	0x10, 0x11, 0x12, 0x13, 0x3C, 0x15, 0x32, 0x26,
	/*18  CN    EM    SB    EC    FS    GS    RS    US */
	0x18, 0x19, 0x3F, 0x27, 0x1C, 0x1D, 0x1E, 0x1F,
	/*20  SP     !     "     #     $     %     &     ' */
	0x40, 0x5A, 0x7F, 0x7B, 0x5B, 0x6C, 0x50, 0x7D,
	/*28   (     )     *     +     ,     -    .      / */
	0x4D, 0x5D, 0x5C, 0x4E, 0x6B, 0x60, 0x4B, 0x61,
	/*30   0     1     2     3     4     5     6     7 */
	0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7,
	/*38   8     9     :     ;     <     =     >     ? */
	0xF8, 0xF9, 0x7A, 0x5E, 0x4C, 0x7E, 0x6E, 0x6F,
	/*40   @     A     B     C     D     E     F     G */
	0x7C, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7,
	/*48   H     I     J     K     L     M     N     O */
	0xC8, 0xC9, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6,
	/*50   P     Q     R     S     T     U     V     W */
	0xD7, 0xD8, 0xD9, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6,
	/*58   X     Y     Z     [     \     ]     ^     _ */
	0xE7, 0xE8, 0xE9, 0xAD, 0xE0, 0xBD, 0x5F, 0x6D,
	/*60   `     a     b     c     d     e     f     g */
	0x79, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	/*68   h     i     j     k     l     m     n     o */
	0x88, 0x89, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96,
	/*70   p     q     r     s     t     u     v     w */
	0x97, 0x98, 0x99, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6,
	/*78   x     y     z     {     |     }     ~    DL */
	0xA7, 0xA8, 0xA9, 0xC0, 0x4F, 0xD0, 0xA1, 0x07,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F,
	0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0x3F, 0xFF
};







// dlist for keeping disk private data areas
dlist_t  Disk_PrivateData_List=NULL;
#define DISK_PDATA_TAG  0x08080808





/*
 *  A freespace has a name that has 3 parts to it.
 *  (1) a parent object name like hda
 *  (2) the word freespace
 *  (3) a number used to keep freespace seg names different
 *
 *  Examples:  hda_freespace1
 *             hda_freespace2
 *             dasdh3_freespace4
 *
 *  This routine returns the number found at the end
 *  of the freespace segment name, ignoring any differences
 *  that may be found in other parts of the name.
 *
 *  Returns: success: integer > 0
 *           failure: -1
 *
 */
static int get_freespace_number(DISKSEG *freespace)
{
	int number=-1;
	int i;

	if (freespace) {

		if (freespace->name) {

			// pretty liberal here...
			// minimum string has length of 2 ... freespac[eN]
			if (strlen(freespace->name)>=2) {

				// walk backwards
				for (i=strlen(freespace->name)-1; i>0; i--) {

					// first e we find ends the search
					if ( freespace->name[i-1] == 'e' ) {

						// convert ascii number to integer
						return(  atoi( &freespace->name[i] ) );

					}

				}

			}

		}

	}

	return number;
}


/*
 *  Called by the get_name_for_disk_segment() routine to come up
 *  with the next available number to use when naming freespace
 *  segments on this drive.
 */
static int  get_next_avail_freespace_number(LOGICALDISK *ld)
{
	int freespace_number=0;
	int rc;
	DISKSEG *seg;

	LOGENTRY();

	rc = GoToStartOfList( ld->parent_objects );
	if ( rc == DLIST_SUCCESS) {

		rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE, (void **) &seg );
		while ( rc == DLIST_SUCCESS ) {

			if (seg->data_type == FREE_SPACE_TYPE) {

				if ( get_freespace_number(seg) > freespace_number ) {
					freespace_number = get_freespace_number(seg);
				}

			}

			rc = GetNextObject( ld->parent_objects,SEGMENT_TAG,(void **)&seg );
		}

	}

	++freespace_number;

	LOGEXIT();
	return freespace_number;
}


/*
 *  Called by insert_DiskSeg_Into_List() to get a name
 *  for the segment storage object being created.
 */
static int  get_name_for_disk_segment( DISKSEG *seg )
{
	int                 rc = EINVAL;
	LOGICALDISK        *ld  = get_logical_disk(seg);
	DISK_PRIVATE_DATA  *disk_pdata;
	SEG_PRIVATE_DATA   *pdata;

	LOGENTRY();

	if (ld) {

		disk_pdata = get_s390_disk_private_data(ld);
		pdata      = (SEG_PRIVATE_DATA *)seg->private_data;

		if (seg->data_type == DATA_TYPE) {

			sprintf(seg->name, "%s%d", ld->name, pdata->minor );
			rc = 0;

		}
		else if (seg->data_type == FREE_SPACE_TYPE) {

			sprintf(seg->name, "%s_freespace%d", ld->name, get_next_avail_freespace_number(ld) );
			rc = 0;

		}
		else if (seg->data_type == META_DATA_TYPE) {

			if (seg->size == ld->size) {
				sprintf(seg->name, "%s_unformatted_space", ld->name );
			}
			else {
				sprintf(seg->name, "%s_metadata", ld->name );
			}
			rc = 0;

		}
		else {
			LOG_ERROR("segment has unknown data type (type=%d)\n", seg->data_type );
		}

	}

	LOGEXITRC();
	return rc;
}


/*
 *  Called to allocate a segment storage object
 */
DISKSEG *  allocate_s390_disk_segment( LOGICALDISK *ld )
{
	int   rc;
	void *handle;
	DISKSEG            *seg=NULL;


	LOGENTRY();


	rc = S390EngFuncs->allocate_segment( NULL, &seg );
	if (rc==0) {

		// insert logical disk into new segment's child dlist
		rc = InsertObject ( (dlist_t)          seg->child_objects,
				    (void *)           ld,
				    (TAG)              DISK_TAG,
				    (void *)           NULL,
				    (Insertion_Modes)  AppendToList,
				    (BOOLEAN)          TRUE,
				    (void **)         &handle );

		if (rc == 0) {

			seg->plugin      = s390_plugin_rec;
			seg->object_type = SEGMENT;
			seg->data_type   = DATA_TYPE;

			memcpy(&seg->geometry, &ld->geometry, sizeof(geometry_t));

			seg->private_data = calloc(1, sizeof(SEG_PRIVATE_DATA));
			if (seg->private_data) {

				((SEG_PRIVATE_DATA *)seg->private_data)->signature    = S390_SEG_MGR_PDATA_SIGNATURE;
				((SEG_PRIVATE_DATA *)seg->private_data)->logical_disk = ld;

			}
			else {
				LOG_ERROR("call to malloc segment private storage area failed\n");
				S390EngFuncs->free_segment( seg );
				seg = NULL;
			}
		}
		else {
			LOG_ERROR("call to insert DISK storage object in segment child_objects dlist failed, RC= %d\n", rc );
		}

	}
	else {
		LOG_ERROR("call to engine_allocate_segment failed, RC= %d\n", rc);
		seg = NULL;
	}

	LOGEXIT();

	return seg;
}


/*
 * Called to free a segment storage object
 */
void free_s390_disk_segment( DISKSEG *seg )
{
	SEG_PRIVATE_DATA  *pdata=NULL;

	LOGENTRY();

	if (seg) {

		LOG_DEBUG("segment name= %s\n", seg->name);

		pdata = (SEG_PRIVATE_DATA *) seg->private_data;

		S390EngFuncs->free_segment( seg );

		if (pdata) free(pdata);
	}

	LOGEXIT();

}





/*
 *  Called to allocate a DISK_PRIVATE_DATA struct and link it into the
 *  list we maintain for disks we manage disk segments on.
 */
int create_s390_disk_private_data( LOGICALDISK *ld )
{
	int rc=ENOMEM;
	DISK_PRIVATE_DATA *disk_pdata;
	void *handle;

	LOGENTRY();

	// make sure dlist is created
	if (Disk_PrivateData_List == NULL) {

		Disk_PrivateData_List = CreateList();

		if (Disk_PrivateData_List == NULL) {
			LOGEXITRC();
			return rc;
		}

	}

	if ( get_s390_disk_private_data(ld ) == NULL ) {

		disk_pdata = (DISK_PRIVATE_DATA *) calloc(1, sizeof(DISK_PRIVATE_DATA) );

		if ( disk_pdata ) {

			disk_pdata->signature          = S390_SEG_MGR_PDATA_SIGNATURE;
			disk_pdata->key                = ld;

			disk_pdata->vsectors_per_block = ld->geometry.bytes_per_sector >> EVMS_VSECTOR_SIZE_SHIFT;
			disk_pdata->geometry.cylinders = ld->geometry.cylinders;
			disk_pdata->geometry.heads     = ld->geometry.heads;
			disk_pdata->geometry.sectors   = ld->geometry.sectors_per_track;
			disk_pdata->geometry.start     = 0;


			rc = InsertObject ( (dlist_t)          Disk_PrivateData_List,
					    (void *)           disk_pdata,
					    (TAG)              DISK_PDATA_TAG,
					    (void *)           NULL,
					    (Insertion_Modes)  AppendToList,
					    (BOOLEAN)          TRUE,
					    (void **)         &handle );

			if ( rc ) {
				free(disk_pdata);
			}

		}
		else {
			rc = ENOMEM;
		}
	}
	else {
		rc = 0;
	}

	LOGEXITRC();
	return rc;
}

/*
 *  Called to delete the specified disk private data by pruning it from the double
 *  linked list.
 */
int delete_s390_disk_private_data( LOGICALDISK *ld )
{
	int                rc = EINVAL;
	DISK_PRIVATE_DATA *disk_pdata = get_s390_disk_private_data(ld );


	LOGENTRY();

	// make sure dlist exists
	if ( Disk_PrivateData_List == NULL ) {
		LOGEXITRC();
		return rc;
	}

	// delete object from list and free its memory
	if (disk_pdata) {

		DeleteObject( Disk_PrivateData_List, disk_pdata );

		free(disk_pdata);

		rc = 0;
	}
	else {
		rc = EINVAL;
	}

	LOGEXITRC();
	return rc;
}


/*
 *  Returns the private data area for the specified disk.
 */
DISK_PRIVATE_DATA * get_s390_disk_private_data( LOGICALDISK *ld )
{
	int                 rc;
	DISK_PRIVATE_DATA  *disk_pdata = NULL;


	LOGENTRY();

	// make sure dlist is created
	if ( Disk_PrivateData_List ) {

		rc = GoToStartOfList( Disk_PrivateData_List );
		if ( rc == DLIST_SUCCESS ) {

			rc = GetObject( Disk_PrivateData_List,DISK_PDATA_TAG,NULL,TRUE,(void **)&disk_pdata );

			while (rc == DLIST_SUCCESS) {

				// compare storage object ptr to key field in private data area
				if ( disk_pdata->key == ld ) {
					LOGEXIT();
					return disk_pdata;

				}

				rc = GetNextObject( Disk_PrivateData_List,DISK_PDATA_TAG,(void **) &disk_pdata );

			};

		}

	}

	LOG_ERROR("error: get_disk_private_data(%s) returning NULL\n", ld->name);
	return NULL;
}



/*
 *  Called to see if the two disks LD1 and LD2 are the same
 *  physical disks by comapring geometry and volume labels.
 *
 *  Returns:  TRUE if they are the same physical disk
 *            FALSE if not
 */
static BOOLEAN same_physical_disk( LOGICALDISK *ld1, LOGICALDISK *ld2 )
{
	int rc=0,blocksize;

	BOOLEAN            same_disk_result=FALSE;
	lba_t              lba;

	char               volume_label_sector1[EVMS_VSECTOR_SIZE];
	volume_label_t    *vlabel = (volume_label_t *) volume_label_sector1;

	char               volume_label_sector2[EVMS_VSECTOR_SIZE];
	volume_label_t    *vlabel2 = (volume_label_t *) volume_label_sector2;

	char               saved_volid[VOLSER_LENGTH];

	struct hd_geometry geo1;
	struct hd_geometry geo2;



	LOGENTRY();

	/*
	 *  test for same geometry
	 */
	geo1.cylinders = ld1->geometry.cylinders;
	geo1.heads     = ld1->geometry.heads;
	geo1.sectors   = ld1->geometry.sectors_per_track;
	geo1.start     = 0;

	geo2.cylinders = ld2->geometry.cylinders;
	geo2.heads     = ld2->geometry.heads;
	geo2.sectors   = ld2->geometry.sectors_per_track;
	geo2.start     = 0;

	blocksize      = ld1->geometry.bytes_per_sector >> EVMS_VSECTOR_SIZE_SHIFT;

	if ( memcmp( &geo1, &geo2, sizeof(struct hd_geometry) )!=0 ) {
		LOG_DEBUG("disk geometry is not the same for both disks\n");
		LOGEXIT();
		return FALSE;
	}

	/*
	 *  read the volume label record off disk 1
	 */
	lba  = VOLUME_LABEL_BLOCK_NUMBER * blocksize;

	rc = EVMS_READ( ld1, lba, 1, (void *) vlabel);
	if (rc) {
		LOG_DEBUG("error(%d) reading volume label off %s.\n",rc, ld1->name);
		ld1->flags |= SOFLAG_CORRUPT;
		LOGEXIT();
		return FALSE;
	}

	/*
	 *  set test pattern in volid field
	 */
	memcpy( saved_volid, vlabel->volid, 6 );
	memcpy( vlabel->volid, EVMS_MP_TEST_PATTERN, 6 );

	rc = EVMS_WRITE( ld1, lba, 1, (void *) vlabel);
	if (rc) {
		LOG_ERROR("error, io error writing multipath test pattern to disk\n");
		ld1->flags |= SOFLAG_CORRUPT;
		LOGEXIT();
		return FALSE;
	}

	/*
	 * read volume label off disk 2 and see if we get a match on
	 * our test pattern.
	 */
	rc = EVMS_READ( ld2, lba, 1, (void *) vlabel2);
	if (rc == 0) {

		if ( memcmp(vlabel2->volid, EVMS_MP_TEST_PATTERN, 6) == 0 ) {
			same_disk_result = TRUE;
		}

	}

	/*
	 * restore the original volume label on disk 1
	 */
	memcpy( vlabel->volid, saved_volid, 6 );
	rc = EVMS_WRITE( ld1, lba, 1, (void *) vlabel);
	if (rc) {
		rc = EVMS_WRITE( ld1, lba, 1, (void *) vlabel);
		if (rc) {
			LOG_ERROR("error, io error restoring multipath test pattern to disk\n");
			ld1->flags |= SOFLAG_CORRUPT;
		}
	}


	LOGEXIT();
	return same_disk_result;
}



/*
 *  Called to see if the specified disk (LD1) is actually
 *  a multipath to an existing disk storage object (LD2).
 *
 *  Proceedure:
 *
 *  Pick first LD2 from our private logical disk list
 *
 *  IF  ... volume id of LD1 matches the volume id of LD2
 *  THEN ... {
 *              modify volume label on disk LD1 using predetermined test pattern
 *              read volume label off disk LD2
 *              IF ... the volume id from disk LD2 is our test pattern
 *              THEN ... LD1 and LD2 are the same disk, return LD2 to caller
 *              ELSE ... try another LD2 by continuing to next list item
 *           }
 *  ELSE ... try another LD2 by continuing to next list item
 *
 */
LOGICALDISK * get_s390_multipath_disk( LOGICALDISK *ld1, volume_label_t *vlabel )
{
	int                rc;
	LOGICALDISK       *ld2=NULL;
	DISK_PRIVATE_DATA *disk_pdata=NULL;


	LOGENTRY();

	if ( Disk_PrivateData_List ) {

		rc = GoToStartOfList( Disk_PrivateData_List );
		if ( rc == DLIST_SUCCESS ) {

			rc = GetObject( Disk_PrivateData_List,DISK_PDATA_TAG,NULL,TRUE,(void **)&disk_pdata );

			while (rc == DLIST_SUCCESS) {

				ld2 = (LOGICALDISK *)disk_pdata->key;

				if ( memcmp( &vlabel->volid, &disk_pdata->vlabel.volid, 6) == 0) {

					if ( same_physical_disk( ld1, ld2 ) == TRUE) {
						LOGEXIT();
						return ld2;
					}

				}

				rc = GetNextObject( Disk_PrivateData_List,DISK_PDATA_TAG,(void **) &disk_pdata );
			};

		}

	}

	LOGEXIT();
	return NULL;
}



/*
 *   Returns the DISKSEG struct ptr to the freespace seg following the
 *   specified DISKSEG.  Returns NULL if freespace is not found.
 *
 *  Consider the following seglist ...
 *
 *    Seg_1     Seg2       Seg3       Seg4      Seg5
 *    MBR_SEG   DATA_SEG   Free_Seg   EBR_SEG   DATA_SEG
 *
 *   If we were asked for the freespace seg following Seg2 we would
 *   respond by returning Seg3.  The freespace seg must immediately
 *   follow the specified DISKSEG.
 *
 */
DISKSEG * get_freespace_following_s390_segment(DISKSEG *seg)
{
	dlist_t      seglist=NULL;
	DISKSEG     *prev = NULL;
	DISKSEG     *this = NULL;
	int          rc;
	LOGICALDISK  *ld=NULL;


	// get logical disk and the entire segment list for the disk
	ld = get_logical_disk( seg );
	if ( ld ) {
		seglist = ld->parent_objects;
	}

	if ( (ld) && (seglist) ) {

		rc = GoToStartOfList( seglist );
		if (rc)	return NULL;

		rc = GetObject( seglist,SEGMENT_TAG,NULL,TRUE,(void **) &prev );
		if (rc)	return NULL;

		do {

			rc = GetNextObject( seglist,SEGMENT_TAG, (void **) &this );
			if (rc)	return NULL;

			if (this != NULL) {

				if (prev == seg) {

					if ( this->data_type==FREE_SPACE_TYPE ) {
						return this;
					}
					else {
						return NULL;
					}

				}

				prev=this;
			}

		} while (this != NULL);

		return NULL;
	}
	else {
		return NULL;
	}

}



/*
 *   Called to try and find free space between disk data segments, returning a
 *   new free space DISKSEG struct if a gap between existing segments is found.
 */

DISKSEG * find_freespace_in_seglist( dlist_t seglist )
{
	DISKSEG            *prev = NULL;
	DISKSEG            *this = NULL;
	DISKSEG            *freeseg  = NULL;
	LOGICALDISK        *ld;
	int                 rc;
	int64_t             gap;


	LOGENTRY();

	if (seglist) {

		rc = GoToStartOfList( seglist );
		if (rc == DLIST_SUCCESS) {

			rc = GetObject( seglist,SEGMENT_TAG,NULL,TRUE,(void **) &prev );
			if (rc == DLIST_SUCCESS) {

				// need disk private data area
				ld = get_logical_disk( prev );
				if (ld) {

					do {

						rc = GetNextObject( seglist,SEGMENT_TAG, (void **) &this );
						if ( rc == DLIST_SUCCESS) {

							gap = this->start - ( prev->start + prev->size );

							if ( gap > 0 ) {

								// allocate a segment storage object from the engine
								freeseg = allocate_s390_disk_segment(ld);

								if (freeseg ) {

									freeseg->data_type    = FREE_SPACE_TYPE;
									freeseg->size         = gap;
									freeseg->start        = prev->start + prev->size;

									freeseg->flags       &= ~SOFLAG_DIRTY;

									LOGEXIT();
									return freeseg;

								}
								else {
									LOGEXIT();
									return NULL;
								}

							}
							else {
								prev = this;
							}
						}

					} while ( rc==DLIST_SUCCESS );
				}
			}
		}
	}


	LOGEXIT();
	return NULL;
}


/*
 *  Called to look for free space on the logical disk and to create
 *  segments which will expose the free space on the disk.
 */
int  find_freespace_on_s390_disk( LOGICALDISK *ld )
{
	DISKSEG            *freeseg              = NULL;
	DISKSEG            *seg                  = NULL;
	DISKSEG            *segaddr              = NULL;
	uint                segcount;
	sector_count_t      sectors_left_on_disk = 0;
	int                 rc;
	lba_t               freespace_start_lba;
	DISK_PRIVATE_DATA  *disk_pdata;
	dlist_t             seglist = ld->parent_objects;



	LOGENTRY();

	// need disk private data area
	disk_pdata = get_s390_disk_private_data( ld );
	if (disk_pdata==NULL) return EINVAL;

	rc = GetListSize(seglist, &segcount);
	if (rc)	segcount = 0;

	if ( segcount > 0 ) {					   /* IF ... we have disk segments */

		do {						       /* THEN ... look for gaps between them */
			seg = find_freespace_in_seglist( seglist );

			if (seg != NULL) {
				
				segaddr = insert_s390_segment_into_list( seglist, seg);
				if (segaddr==NULL) {
					free_s390_disk_segment( seg );
					rc = ENOMEM;
					LOGEXITRC();
					return rc;					
				}

			}

		} while (seg != NULL);


		rc = GoToEndOfList( seglist );	     /* AND ... calc number of free sectors at the */
		if (rc) {
			LOGEXIT();
			return 0;			 /*         end of disk after the last segment */
		}

		rc = GetObject(seglist,SEGMENT_TAG,NULL,TRUE,(ADDRESS *) &seg );
		if (rc) {
			LOGEXIT();
			return 0;
		}

		sectors_left_on_disk = ld->size - ( seg->start + seg->size );

		if ( sectors_left_on_disk > 0 ) {
			freespace_start_lba  =  seg->start + seg->size;
		}
		else {
			sectors_left_on_disk = 0;
			freespace_start_lba  = 0;
		}
	}
	else {	// there are no segments on the disk at all so just create a
		// single free space segment for the disk!
		freespace_start_lba  = 0;
		sectors_left_on_disk = ld->size;
	}

	if ( sectors_left_on_disk > 0 ) {

		freeseg = allocate_s390_disk_segment(ld);
		if (freeseg) {

			freeseg->data_type  = FREE_SPACE_TYPE;
			freeseg->size       = sectors_left_on_disk;
			freeseg->start      = freespace_start_lba;

			freeseg->flags     &= ~SOFLAG_DIRTY;

			segaddr = insert_s390_segment_into_list( seglist, freeseg);
			if (segaddr==NULL) {

				int i;

				for (i=1; i<10; i++) {

					segaddr = insert_s390_segment_into_list( seglist, freeseg);
					if (segaddr==NULL) {
						LOG_DEBUG("error, insert_DiskSeg_Into_List returned an error\n");
						free_s390_disk_segment(freeseg);
						LOGEXIT();
						return EIO;
					}
					else {
						break;
					}
				}

			}

		}
		else {
			LOGEXIT();
			return EIO;
		}
	}

	// lastly, merge adjacent free areas
	merge_adjacent_freedisksegs_in_list( ld->parent_objects );

	LOGEXIT();
	return 0;
}





/*
 *   Called to try and join neighboring free segments together.
 *
 *   NOTE:  Assumes that the DISKSEG structs are kept in an ordered list
 *
 */
static int  merge_freespace_segments( dlist_t seglist )
{
	DISKSEG            *prev;
	DISKSEG            *this;
	int                 rc;
	DISK_PRIVATE_DATA  *disk_pdata;
	LOGICALDISK        *ld;


	LOGENTRY();

	rc = GoToStartOfList( seglist );
	if ( rc == DLIST_SUCCESS) {

		rc = GetObject( seglist,SEGMENT_TAG,NULL,TRUE, (void **) &prev );
		if ( rc == DLIST_SUCCESS ) {

			// need disk private data area
			ld         = get_logical_disk(prev);
			disk_pdata = get_s390_disk_private_data( ld );
			if (disk_pdata==NULL) return EINVAL;

			do {

				rc = GetNextObject( seglist,SEGMENT_TAG, (void **) &this );

				if (rc==DLIST_SUCCESS) {

					if ( ( prev==NULL ) ||
					     ( this->data_type != FREE_SPACE_TYPE )||
					     ( prev->data_type != FREE_SPACE_TYPE )) {
						prev = this;
						continue;
					}

					if (get_freespace_number(prev) > get_freespace_number(this) ) {
						rc = DeleteObject( seglist, prev );
						if (rc == DLIST_SUCCESS) {
							this->start -= prev->size;
							this->size  += prev->size;
							free_s390_disk_segment( prev );
							LOG_DEBUG("        kept seg: %s  start: %"PRIu64"  size: %"PRIu64"\n", this->name, this->start, this->size );
							return DLIST_SUCCESS;
						}
						else {
							return rc;
						}
					}
					else {
						rc = DeleteObject( seglist, this );
						if (rc == DLIST_SUCCESS) {
							prev->size  += this->size;
							free_s390_disk_segment( this );
							LOG_DEBUG("        kept seg: %s  start: %"PRIu64"  size: %"PRIu64"\n", prev->name, prev->start, prev->size );
							return DLIST_SUCCESS;
						}
						else {
							return rc;
						}
					}

				}

			} while (rc==DLIST_SUCCESS);  /* continue till we hit the end of the list */
		}
	}

	rc = DLIST_END_OF_LIST;

	LOGEXITRC();
	return rc;
}





/*
 *   Called to try and join neighboring free segments together.
 *
 *   NOTE:  Assumes that the DISKSEG structs are kept in an ordered list
 *
 */
int  merge_adjacent_freedisksegs_in_list( dlist_t seglist )
{
	int         rc;

	LOGENTRY();

	do {

		rc = merge_freespace_segments(seglist);

	} while (rc==DLIST_SUCCESS);  /* continue while there are segs to merge */


	rc = 0;

	LOGEXITRC();
	return rc;
}


/*
 *  Wrapper routine that is called to remove a segment storage object from a dlist.
 */
int  remove_s390_segment_from_list( dlist_t seglist, storage_object_t *seg )
{
	int                rc;


	LOGENTRY();
	LOG_DEBUG("segment name= %s\n",  seg->name );

	rc = DeleteObject( seglist, seg );

	if (rc == DLIST_SUCCESS) {

		S390EngFuncs->unregister_name( seg->name );

	}
	else {
		LOG_ERROR("call to DeleteObject failed\n");
	}

	LOGEXITRC();

	return rc;
}


/*
 *   Called to name a new disk segment and insert it into the
 *   ordered segment list for the logical disk.
 *
 *   Returns the dlist handle if successful.
 */
void *  insert_s390_segment_into_list( dlist_t seglist, storage_object_t *seg)
{
	int    rc=EINVAL;
	void  *result=NULL;


	LOGENTRY();
	LOG_DEBUG("seg start= %"PRIu64"   size= %"PRIu64"\n", seg->start, seg->size);


	rc = get_name_for_disk_segment( seg );
	if (rc) {
		LOGEXIT();
		return NULL;
	}

	rc = S390EngFuncs->register_name( seg->name );
	if (rc == 0) {
		result = insert_s390_segment_into_ordered_list( seglist, seg );
	}
	else {
		LOG_ERROR("error, get_name_for_disk_segment failed, RC= %d\n", rc);
		LOGEXIT();
		return NULL;
	}


	LOG_DEBUG("returning %p\n", result);
	LOGEXIT();
	return result;
}


/*
 *   Called to insert a disk segment into a segment list that we
 *   are trying to keep ordered by the segments starting LBA
 *   number.
 *
 *   Returns the dlist handle if successful.
 */
void *  insert_s390_segment_into_ordered_list( dlist_t seglist, storage_object_t *seg )
{
	int                rc;
	void              *handle=NULL;
	storage_object_t  *seg2;
	lba_t              seg2_end_lba;
	BOOLEAN            overlapping=FALSE;


	LOGENTRY();
	LOG_DEBUG("seg name= %s   seg start= %"PRIu64"  ends= %"PRIu64"  size= %"PRIu64"\n", seg->name, seg->start, seg->start+seg->size-1, seg->size );


	rc = GoToStartOfList( seglist );
	if (rc == DLIST_SUCCESS) {

		while (rc == DLIST_SUCCESS) {

			rc = GetObject( seglist, SEGMENT_TAG,NULL,TRUE, (ADDRESS *) &seg2 );

			if (rc == DLIST_SUCCESS) {

				seg2_end_lba = seg2->start + seg2->size - 1;

				// test and set ... overlapping segments flag
				if ( (  seg->start >= seg2->start )&&
				     (  seg->start <= seg2_end_lba)) {
					overlapping = TRUE;
				}
				else if ( ( seg->start  <  seg2->start ) &&
					  ( seg2->start <= (seg->start + seg->size - 1)) ) {
					overlapping = TRUE;
				}
				else {
					overlapping = FALSE;
				}

				if ( overlapping == TRUE ) {

					LOG_DEBUG("Error ... Overlapping Segments ...\n");
					LOG_DEBUG("seg2:   name: %s\n", seg2->name );
					LOG_DEBUG("       start: %"PRIu64"\n", seg2->start );
					LOG_DEBUG("        size: %"PRIu64"\n", seg2->size );
					LOG_DEBUG("         end: %"PRIu64"\n", seg2_end_lba );
					LOG_DEBUG(" overlap lba: %"PRIu64"\n", seg->start );

					rc = EINVAL;	// must be genuine partition overlap
					break;		// break out of loop ... looking for insertion pt

				}


				// test for ... valid insertion point
				if (seg2->start > seg->start ) break;

				rc = NextItem( seglist );
			}

		}

	}


	switch (rc) {
	
	case DLIST_SUCCESS:	 /* Ok, found a segment we should insert in front of */

		rc = InsertObject ( (dlist_t)          seglist,
				    (void *)           seg,
				    (TAG)              SEGMENT_TAG,
				    (void *)           NULL,
				    (Insertion_Modes)  InsertBefore,
				    (BOOLEAN)          TRUE,
				    (void **)         &handle );
		break;

	case DLIST_EMPTY:
	case DLIST_END_OF_LIST:	 /* This new segment lays out on the disk at a higher */
		/* LBA than any other existing segment.  So, just    */
		/* insert at end of the segment list.                */

		rc = InsertObject ( (dlist_t)          seglist,
				    (void *)           seg,
				    (TAG)              SEGMENT_TAG,
				    (void *)           NULL,
				    (Insertion_Modes)  AppendToList,
				    (BOOLEAN)          TRUE,
				    (void **)         &handle );
		break;

	default:     /* REAL ERROR ... return NULL ptr */
		LOG_ERROR("error, insertion failed ... RC= %d\n",  rc);
		break;
	}


	LOGEXIT();

	if (rc) {
		return NULL;
	}
	else {
		return handle;
	}
}



/*
 *  Called to add a segment to the logical disk's segment DLIST. First, convert the
 *  partition record to a DISKSEG struct.  Then, insert the DISKSEG struct into
 *  the DLIST hanging off the logical disk.
 */
DISKSEG * create_s390_metadata_segment( LOGICALDISK  *ld,  lba_t start, sector_count_t size )
{
	DISKSEG *metadata=NULL;

	LOGENTRY();

	metadata = allocate_s390_disk_segment( ld );
	if (metadata) {

		metadata->size        = size;
		metadata->start       = start;
		metadata->data_type   = META_DATA_TYPE;

	}

	LOGEXIT();
	return metadata;
}


/*
 *  Convert from ascii string to ebcdic string
 */
void  atoe( char *buffer, int length )
{
	int i;

	if ( buffer ) {

		for (i=0; i<length; i++) {

			buffer[i] = (char) ASCtoEBC[ (int) buffer[i] ];

		}

	}

}


/*
 *  Convert from ebcdic string to ascii string
 */
void  etoa( char *buffer, int length )
{
	int i;

	if ( buffer ) {

		for (i=0; i<length; i++) {

			buffer[i] = (char) EBCtoASC[ (int) buffer[i] ];

		}

	}

}


/*
 *  Called to build an ebcdic dataset name for the specified
 *  evms segment. The name will be stored in the format 1
 *  dscb that hangs off the segment private data area.
 */
static int  get_dataset_name_for_disk_segment( LOGICALDISK *ld, DISKSEG *seg )
{
	int                rc=EINVAL;
	SEG_PRIVATE_DATA  *pdata=NULL;
	char               dsname[MAX_DSNAME_LENGTH+1];
	char               volid[VOLSER_LENGTH+1];
	DISK_PRIVATE_DATA *disk_pdata=NULL;


	LOGENTRY();

	if (ld && seg) {

		disk_pdata = get_s390_disk_private_data( ld );
		pdata      = (SEG_PRIVATE_DATA *)seg->private_data;

		if (disk_pdata && pdata) {

			// need volume id field from the dasd volume label
			memcpy(volid, &disk_pdata->vlabel.volid, VOLSER_LENGTH);
			etoa(volid,VOLSER_LENGTH);
			volid[VOLSER_LENGTH]=0x00;

			// use same format as fdasd
			sprintf( dsname, "%s%s%s%04X%s","LINUX.V", volid, ".PART", pdata->minor, ".NEW" );
			memset( &pdata->f1.DS1DSNAM, 0x20, MAX_DSNAME_LENGTH );
			memcpy( &pdata->f1.DS1DSNAM, dsname, strlen(dsname) );
			atoe( pdata->f1.DS1DSNAM, MAX_DSNAME_LENGTH );
			pdata->f1.DS1DSNAM[MAX_DSNAME_LENGTH]=0x00;

			rc = 0;

		}

	}

	LOGEXITRC();
	return rc;
}




/*
 *  Initialize a new format 1 dscb for callers data segment
 */
int init_format1_dscb( LOGICALDISK *ld, DISKSEG *seg )
{
	int                 rc=EINVAL;
	time_t              current_time = time(NULL);
	struct tm          *today=NULL;
	DISK_PRIVATE_DATA  *disk_pdata = NULL;
	SEG_PRIVATE_DATA   *pdata=NULL;


	LOGENTRY();

	if (ld && seg) {

		disk_pdata = get_s390_disk_private_data(ld);
		pdata      = (SEG_PRIVATE_DATA *) seg->private_data;

		if (disk_pdata && pdata) {

			memset( &pdata->f1, 0, sizeof(format1_label_t) );

			get_dataset_name_for_disk_segment( ld, seg );

			memcpy( pdata->f1.DS1SYSCD, S390_SYSTEM_CODE, S390_SYSTEM_CODE_LENGTH );
			atoe( pdata->f1.DS1SYSCD, S390_SYSTEM_CODE_LENGTH );

			memcpy(&pdata->f1.DS1DSSN, &disk_pdata->vlabel.volid, S390_DATA_SET_SN_LENGTH);

			today = gmtime(&current_time);		       // get time

			pdata->f1.DS1CREDT.year  = today->tm_year;     // creation date
			pdata->f1.DS1CREDT.day   = today->tm_yday;

			pdata->f1.DS1EXPDT.year  = 99;		       // expiration date
			pdata->f1.DS1EXPDT.day   = 365;

			memcpy( &pdata->f1.DS1REFD, &pdata->f1.DS1CREDT, sizeof( labeldate_t) );  // last referenced date

			pdata->f1.DS1FMTID = 0xf1;
			pdata->f1.DS1VOLSQ = 1;
			pdata->f1.DS1NOEPV = 1;
			pdata->f1.DS1RECFM = 0x88;
			pdata->f1.DS1BLKL  = disk_pdata->vsectors_per_block;
			pdata->f1.DS1LRECL = disk_pdata->vsectors_per_block;
			pdata->f1.DS1DSIND = 0x80;
			pdata->f1.DS1SCAL1 = 0x80;

			pdata->f1.DS1EXT1.typeind = 0x81;

			rc = 0;
		}
	}

	LOGEXITRC();
	return rc;
}


/*
 *  Initialize a format 5 dscb for callers logical disk
 */
int init_format5_dscb( LOGICALDISK *ld )
{
	int                 rc=EINVAL;
	DISK_PRIVATE_DATA  *disk_pdata = NULL;
	DISKSEG            *seg;
	format5_label_t    *f5;
	int                 extent_number = 0;
	u_int16_t           rta=0;
	u_int16_t           fc=0;
	u_int16_t           ft=0;


	LOGENTRY();

	if (ld) {

		disk_pdata = get_s390_disk_private_data(ld);

		if (disk_pdata) {

			f5 = &disk_pdata->f5;

			memset(f5, 0, sizeof(format5_label_t));	     // clear it
			memset(&f5->DS5KEYID, 0x05, 4 );	     // set key field
			f5->DS5FMTID = 0xF5;			     // set format 5 identifier

			rc = GoToStartOfList( ld->parent_objects );  // walk disk segment list finding all freespace segs
			if ( rc == DLIST_SUCCESS) {		     // and add them as extents

				rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE, (void **) &seg );
				while ( rc == DLIST_SUCCESS ) {

					if (seg->data_type == FREE_SPACE_TYPE) {

						LOG_DEBUG("adding segment %s to descriptor\n", seg->name);

						rta = (seg->start/disk_pdata->vsectors_per_block) / disk_pdata->geometry.sectors;
						fc  = (seg->size/disk_pdata->vsectors_per_block) / (disk_pdata->geometry.heads*disk_pdata->geometry.sectors);
						ft  = (seg->size/disk_pdata->vsectors_per_block) % (disk_pdata->geometry.heads*disk_pdata->geometry.sectors);
						ft  = ft / disk_pdata->geometry.sectors;

						if (extent_number == 0) {	 // use 1st available extent
							f5->DS5AVEXT.t  = rta;
							f5->DS5AVEXT.fc = fc;
							f5->DS5AVEXT.ft = ft;
						}
						else if (extent_number < 8) {	 // use next 7 extents
							f5->DS5EXTAV[extent_number-1].t  = rta;
							f5->DS5EXTAV[extent_number-1].fc = fc;
							f5->DS5EXTAV[extent_number-1].ft = ft;
						}
						else if (extent_number < 26) {	 // use next 18 extents
							f5->DS5MAVET[extent_number-8].t  = rta;
							f5->DS5MAVET[extent_number-8].fc = fc;
							f5->DS5MAVET[extent_number-8].ft = ft;
						}
						else {
							LOG_ERROR("error, no more freespace extents available in F5 dscb\n");
						}
						++extent_number;
					}

					rc = GetNextObject( ld->parent_objects,SEGMENT_TAG,(void **)&seg );
				}

			}

			rc = 0;
		}

	}

	LOGEXITRC();
	return rc;
}



/*
 *  Called to fixup names of logial and embedded partitions because their names will
 *  change if the EBR chain is modified by a delete or create.
 */
int   fixup_390_partition_names( LOGICALDISK *ld )
{
	int                rc;
	DISKSEG           *seg=NULL;
	SEG_PRIVATE_DATA  *pdata=NULL;
	int                minor=1;
	char               old_name[EVMS_NAME_SIZE+1];

	LOGENTRY();

	// unregister all partition names and build new names
	rc = GoToStartOfList( ld->parent_objects );
	if (rc == DLIST_SUCCESS) {

		rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE,(void **) &seg );
		while (rc == DLIST_SUCCESS) {

			pdata = (SEG_PRIVATE_DATA *) seg->private_data;

			if ( seg->data_type == DATA_TYPE ) {

				// unregister segment name
				if (strlen(seg->name)) S390EngFuncs->unregister_name( seg->name );

				// update minor info
				pdata->minor = minor;
				++minor;

				strncpy(old_name, seg->name, EVMS_NAME_SIZE);

				// build new evms segment name
				get_name_for_disk_segment( seg );

				// test if name has changed and then mark the object so we
				// can do a dm rename of the kernel object at commit time.
				if (  (seg->flags & SOFLAG_ACTIVE) &&
				      strlen(old_name) > 0  &&
				      strncmp(seg->name, old_name, EVMS_NAME_SIZE)!=0) {

					s390_schedule_dm_rename(old_name,seg); 

				}

				// build new dsname in format 1 dscb
				get_dataset_name_for_disk_segment( ld, seg);

			}

			rc = GetNextObject( ld->parent_objects, SEGMENT_TAG, (void **) &seg );

		}

	}

	// new register all new partition names
	rc = GoToStartOfList( ld->parent_objects );
	if (rc == DLIST_SUCCESS) {

		rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE,(void **) &seg );
		while (rc == DLIST_SUCCESS) {

			pdata = (SEG_PRIVATE_DATA *) seg->private_data;

			if ( seg->data_type == DATA_TYPE ) {

				S390EngFuncs->register_name( seg->name );

			}

			rc = GetNextObject( ld->parent_objects, SEGMENT_TAG, (void **) &seg );
		}

	}

	// throw away errors like ... DLIST_END_OF_LIST ... which are not real errors
	if ( ( rc == DLIST_EMPTY ) || ( rc == DLIST_END_OF_LIST )) {
		rc = DLIST_SUCCESS;
	}

	LOGEXITRC();
	return 0;
}


/*
 *  Called by private function code to see if a format operation is
 *  allowed on the disk upon which the segment object resides.
 *
 *  Check each segment object to see if it is being consumed by
 *  another storage object.  Also, check to see if it is part of
 *  a volume.  Deny the format request till the user cleans up
 *  these situations.
 */
BOOLEAN i_can_format_disk( LOGICALDISK *ld )
{
	DISKSEG  *seg = NULL;
	int       rc;
	uint      count;

	LOGENTRY();

	if (ld) {

		// test if any of the segs are being consumed
		rc = GoToStartOfList( ld->parent_objects );
		if (rc == DLIST_SUCCESS) {

			rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE,(void **) &seg );
			while (rc == DLIST_SUCCESS) {

				count=0;
				GetListSize(seg->parent_objects, &count);
				if (count || seg->volume ) {
					return FALSE;
				}

				rc = GetNextObject( ld->parent_objects, SEGMENT_TAG, (void **) &seg );
			}

			return TRUE;
		}
		else {
			return TRUE;
		}

	}

	LOGEXIT();
	return FALSE;
}


/*
 *  Called ... to toss all existing segment objects in the list.
 *  All objects in the dlist are expected to be our segment objects.
 */
BOOLEAN prune_s390_segments_from_list( ADDRESS   Object,
				       TAG       ObjectTag,
				       ADDRESS   ObjectHandle,
				       ADDRESS   Parameters,
				       BOOLEAN  *FreeMemory,
				       uint     *Error )
{
	*FreeMemory = FALSE;	       // tells dlist not to free any memory
	*Error      = DLIST_SUCCESS;   // tells dlist we were successful

	if ( ((DISKSEG*)Object)->flags & SOFLAG_ACTIVE) {
		S390EngFuncs->dm_deactivate((DISKSEG *) Object);
	}

	free_s390_disk_segment((DISKSEG *) Object);

	return TRUE;		       // tells dlist to prune the object
}



/*
 *  Called to get the segment storage object corresponding to the specified
 *  device minor number.  This information is kept in the private data area
 *  of gpt segment storage objects.
 *
 *  Returns: DISKSEG * if successful
 *           NULL ptr if minor is not found
 */
DISKSEG * get_s390_segment_from_minor( LOGICALDISK *ld, int minor )
{
	int rc;
	DISKSEG    *seg=NULL;

	LOGENTRY();

	rc = GoToStartOfList( ld->parent_objects );
	if ( rc == DLIST_SUCCESS) {

		rc = GetObject( ld->parent_objects,SEGMENT_TAG,NULL,TRUE, (void **) &seg );
		while ( rc == DLIST_SUCCESS ) {

			if (seg->data_type == DATA_TYPE) {

				if ( ((SEG_PRIVATE_DATA *)seg->private_data)->minor == minor ) {
					return seg;
				}

			}

			rc = GetNextObject( ld->parent_objects,SEGMENT_TAG,(void **)&seg );
		}

	}

	LOGEXIT();
	return NULL;
}


/*
 *  Called to obtain the next available device minor
 *  number for the specified drive. This is an UNUSED
 *  minor.
 *
 *  Note: this routine should not be called during discovery
 *        because it expects all segments to be created for
 *        the drive and not in the middle of adding existing
 *        drive partitions.
 *
 *  Returns: device minor number if successful
 *           -1 if not successful
 */
int get_next_s390_minor( LOGICALDISK *ld )
{
	int                 i;
	DISK_PRIVATE_DATA  *disk_pdata=NULL;


	LOGENTRY();

	disk_pdata = get_s390_disk_private_data(ld);
	if (disk_pdata) {

		for (i=1; i<=disk_pdata->max_partitions_allowed; i++) {

			if ( get_s390_segment_from_minor( ld, i)==NULL ) {
				LOGEXIT();
				return i;
			}

		}

	}


	LOGEXIT();
	return -1;
}


